/*
 * Run.h
 *
 *  Created on: 9. 2. 2014
 *      Author: Jan Travnicek
 */

#ifndef _AUTOMATON_RUN_H__
#define _AUTOMATON_RUN_H__

#include <core/multipleDispatch.hpp>
#include <automaton/AutomatonFeatures.h>
#include <string/LinearString.h>
#include <tree/ranked/RankedTree.h>

#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/TA/DFTA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/VisiblyPushdownDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/NPDTA.h>
#include <global/GlobalData.h>

#include <label/FailStateLabel.h>

#include <deque>
#include <algorithm>
#include <iterator>

namespace automaton {

namespace run {

class Run {
	template < class SymbolType, class RankType, class StateType >
	static std::pair < bool, StateType > calculateState ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const std::tree < std::ranked_symbol < SymbolType, RankType > > & node, std::set < unsigned > & occ, unsigned & i );
	template < class SymbolType, class RankType, class StateType >
	static std::pair < bool, std::set < StateType > > calculateStates ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const std::tree < std::ranked_symbol < SymbolType, RankType > > & node, std::set < unsigned > & occ, unsigned & i );
	template < class SymbolType >
	static bool canPop ( const std::deque < SymbolType > & pushdownStore, const std::vector < SymbolType > & pop );

public:
	template < class SymbolType, class StateType >
	static std::tuple < bool, StateType, std::set < unsigned > > calculateState ( const automaton::DFA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string );
	template < class SymbolType, class StateType >
	static std::tuple < bool, std::set < StateType >, std::set < unsigned > > calculateStates ( const automaton::NFA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string );
	template < class SymbolType, class RankType, class StateType >
	static std::tuple < bool, StateType, std::set < unsigned > > calculateState ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const tree::RankedTree < SymbolType, RankType > & tree );
	template < class SymbolType, class RankType, class StateType >
	static std::tuple < bool, std::set < StateType >, std::set < unsigned > > calculateStates ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const tree::RankedTree < SymbolType, RankType > & tree );
	template < class SymbolType, class StateType >
	static std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > calculateState ( const automaton::InputDrivenDPDA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string );
	template < class SymbolType, class StateType >
	static std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > calculateState ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string );
	template < class SymbolType, class EpsilonType, class StateType >
	static std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > calculateState ( const automaton::RealTimeHeightDeterministicDPDA < SymbolType, EpsilonType, StateType > & automaton, const string::LinearString < SymbolType > & string );
	template < class SymbolType, class EpsilonType, class StateType >
	static std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > calculateState ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const string::LinearString < SymbolType > & string );
	template < class SymbolType, class EpsilonType, class StateType >
	static std::tuple < bool, std::set < StateType >, std::set < std::vector < SymbolType > > > calculateStates ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const string::LinearString < SymbolType > & string );
};

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class StateType >
std::tuple < bool, StateType, std::set < unsigned > > Run::calculateState ( const automaton::DFA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	bool res = true;
	unsigned i = 0;
	std::set < unsigned > occurrences;
	StateType state = automaton.getInitialState ( );

	if ( automaton.getFinalStates ( ).count ( state ) )
		occurrences.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( const SymbolType & symbol : string.getContent ( ) ) {
		auto transition = automaton.getTransitions ( ).find ( std::make_pair ( state, symbol ) );

		if ( transition == automaton.getTransitions ( ).end ( ) ) {
			res = false;
			break;
		}

		state = transition->second;
		i++;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occurrences.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occurrences );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class StateType >
std::tuple < bool, std::set < StateType >, std::set < unsigned > > Run::calculateStates ( const automaton::NFA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	bool res = true;
	unsigned i = 0;
	std::set < unsigned > occurrences;
	std::set < StateType > states {
		automaton.getInitialState ( )
	};

	for ( const StateType & state : states )
		if ( automaton.getFinalStates ( ).count ( state ) )
			occurrences.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << states << std::endl;

	for ( const SymbolType & symbol : string.getContent ( ) ) {
		std::set < StateType > next;

		for ( const StateType & state : states ) {
			auto transitions = automaton.getTransitions ( ).find ( std::make_pair ( state, symbol ) );

			if ( transitions == automaton.getTransitions ( ).end ( ) ) continue;

			next.insert ( transitions->second.begin ( ), transitions->second.end ( ) );
		}

		i++;
		states = next;

		for ( const StateType & state : states )
			if ( automaton.getFinalStates ( ).count ( state ) )
				occurrences.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << states << std::endl;
	}

	return std::make_tuple ( res, states, occurrences );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class RankType, class StateType >
std::pair < bool, StateType > Run::calculateState ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const std::tree < std::ranked_symbol < SymbolType, RankType > > & node, std::set < unsigned > & occ, unsigned & i ) {
	std::vector < StateType > states;

	states.reserve ( ( size_t ) node.getData ( ).getRank ( ) );

	unsigned tmp = i;
	i++;

	bool sign = true;

	for ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & child : node.getChildren ( ) ) {
		std::pair < bool, StateType > res = calculateState ( automaton, child, occ, i );

		if ( res.first == false )
			sign = false;
		else
			states.push_back ( res.second );
	}

	if ( !sign ) return std::make_pair ( false, label::FailStateLabel::instance < StateType > ( ) );

	const auto & it = automaton.getTransitions ( ).find ( std::make_pair ( node.getData ( ), states ) );

	if ( it == automaton.getTransitions ( ).end ( ) ) return std::make_pair ( false, label::FailStateLabel::instance < StateType > ( ) );

	StateType state = it->second;

	if ( automaton.getFinalStates ( ).count ( state ) ) occ.insert ( tmp );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	return std::make_pair ( true, state );
}

template < class SymbolType, class RankType, class StateType >
std::tuple < bool, StateType, std::set < unsigned > > Run::calculateState ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const tree::RankedTree < SymbolType, RankType > & tree ) {
	std::set < unsigned > occ;
	unsigned i = 0;
	std::pair < bool, StateType > res = calculateState ( automaton, tree.getContent ( ), occ, i );

	return std::make_tuple ( res.first, res.second, occ );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class RankType, class StateType >
std::pair < bool, std::set < StateType > > Run::calculateStates ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const std::tree < std::ranked_symbol < SymbolType, RankType > > & node, std::set < unsigned > & occ, unsigned & i ) {
	std::vector < std::set < StateType > > resStates;

	resStates.reserve ( ( size_t ) node.getData ( ).getRank ( ) );

	unsigned tmp = i;
	i++;

	bool sign = true;

	for ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & child : node.getChildren ( ) ) {
		std::pair < bool, std::set < StateType > > childStates = calculateStates ( automaton, child, occ, i );

		if ( childStates.second.empty ( ) )
			sign = false;

		resStates.push_back ( childStates.second );
	}

	std::set < StateType > states;

	for ( const auto & transition : automaton.getTransitions ( ) ) {
		if ( transition.first.first != node.getData ( ) )
			continue;

		unsigned rank = ( unsigned ) transition.first.first.getRank ( );
		unsigned j;

		for ( j = 0; j < rank; j++ )
			if ( !resStates[j].count ( transition.first.second[j] ) )
				break;

		if ( j == rank )
			states.insert ( transition.second.begin ( ), transition.second.end ( ) );
	}

	for ( const StateType & state : states )
		if ( automaton.getFinalStates ( ).count ( state ) ) occ.insert ( tmp );

	if ( common::GlobalData::verbose )
		std::clog << states << std::endl;

	return std::make_pair ( sign, states );
}

template < class SymbolType, class RankType, class StateType >
std::tuple < bool, std::set < StateType >, std::set < unsigned > > Run::calculateStates ( const automaton::NFTA < SymbolType, RankType, StateType > & automaton, const tree::RankedTree < SymbolType, RankType > & tree ) {
	std::set < unsigned > occ;
	unsigned i = 0;
	std::pair < bool, std::set < StateType > > res = calculateStates ( automaton, tree.getContent ( ), occ, i );

	return std::make_tuple ( res.first, res.second, occ );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType >
bool Run::canPop ( const std::deque < SymbolType > & pushdownStore, const std::vector < SymbolType > & pop ) {
	unsigned popSize = pop.size ( );
	unsigned pushdownStoreSize = pushdownStore.size ( );

	if ( pushdownStoreSize < popSize ) return false;

	for ( unsigned i = 0; i < popSize; i++ )
		if ( pop[i] != pushdownStore[pushdownStoreSize - i - 1] )
			return false;

	return true;
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class StateType >
std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > Run::calculateState ( const automaton::InputDrivenDPDA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	StateType state = automaton.getInitialState ( );
	std::deque < SymbolType > pushdownStore {
		automaton.getInitialSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( const SymbolType & symbol : string.getContent ( ) ) {
		auto transition = automaton.getTransitions ( ).find ( std::make_pair ( state, symbol ) );

		if ( transition == automaton.getTransitions ( ).end ( ) ) {
			res = false;
			break;
		}

		const std::pair < std::vector < SymbolType >, std::vector < SymbolType > > & operation = automaton.getPushdownStoreOperations ( ).find ( symbol )->second;

		if ( !canPop ( pushdownStore, operation.first ) ) {
			res = false;
			break;
		}

		for ( unsigned j = 0; j < operation.first.size ( ); j++ ) pushdownStore.pop_back ( );

		for ( const auto & push : std::make_reverse ( operation.second ) ) pushdownStore.push_back ( push );

		state = transition->second;
		i++;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class StateType >
std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > Run::calculateState ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	StateType state = automaton.getInitialState ( );
	std::deque < SymbolType > pushdownStore {
		automaton.getBottomOfTheStackSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( const SymbolType & symbol : string.getContent ( ) ) {
		if ( automaton.getCallInputAlphabet ( ).count ( symbol ) ) {
			auto transition = automaton.getCallTransitions ( ).find ( std::make_pair ( state, symbol ) );

			if ( transition == automaton.getCallTransitions ( ).end ( ) ) {
				res = false;
				break;
			}

			pushdownStore.push_back ( transition->second.second );
			state = transition->second.first;
		} else if ( automaton.getReturnInputAlphabet ( ).count ( symbol ) ) {
			auto transition = automaton.getReturnTransitions ( ).find ( std::make_tuple ( state, symbol, pushdownStore.back ( ) ) );

			if ( transition == automaton.getReturnTransitions ( ).end ( ) ) {
				res = false;
				break;
			}

			if ( pushdownStore.back ( ) != automaton.getBottomOfTheStackSymbol ( ) ) pushdownStore.pop_back ( );

			state = transition->second;
		} else if ( automaton.getLocalInputAlphabet ( ).count ( symbol ) ) {
			auto transition = automaton.getLocalTransitions ( ).find ( std::make_pair ( state, symbol ) );

			if ( transition == automaton.getLocalTransitions ( ).end ( ) ) {
				res = false;
				break;
			}

			state = transition->second;
		} else {
			res = false;
			break;
		}

		i++;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class EpsilonType, class StateType >
std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > Run::calculateState ( const automaton::RealTimeHeightDeterministicDPDA < SymbolType, EpsilonType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	StateType state = automaton.getInitialState ( );
	std::deque < SymbolType > pushdownStore {
		automaton.getBottomOfTheStackSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( auto symbolIter = string.getContent ( ).begin ( ); symbolIter != string.getContent ( ).end ( ); ) {

		auto callTransition = automaton.getCallTransitions ( ).find ( std::make_pair ( state, * symbolIter ) );

		if ( callTransition != automaton.getCallTransitions ( ).end ( ) ) {
			symbolIter++;
			i++;
		} else {
			callTransition = automaton.getCallTransitions ( ).find ( std::make_pair ( state, EpsilonType ( ) ) );
		}

		if ( callTransition != automaton.getCallTransitions ( ).end ( ) ) {
			pushdownStore.push_back ( callTransition->second.second );
			state = callTransition->second.first;
		} else {
			res = false;
		}

		if ( !res ) {
			auto returnTransition = automaton.getReturnTransitions ( ).find ( std::make_tuple ( state, * symbolIter, pushdownStore.back ( ) ) );

			if ( returnTransition != automaton.getReturnTransitions ( ).end ( ) ) {
				symbolIter++;
				i++;
			} else {
				returnTransition = automaton.getReturnTransitions ( ).find ( std::make_tuple ( state, EpsilonType ( ), pushdownStore.back ( ) ) );
			}

			if ( returnTransition != automaton.getReturnTransitions ( ).end ( ) ) {
				pushdownStore.pop_back ( );
				state = returnTransition->second;
			} else {
				res = false;
			}
		}

		if ( !res ) {
			auto localTransition = automaton.getLocalTransitions ( ).find ( std::make_pair ( state, * symbolIter ) );

			if ( localTransition != automaton.getLocalTransitions ( ).end ( ) ) {
				symbolIter++;
				i++;
			} else {
				localTransition = automaton.getLocalTransitions ( ).find ( std::make_pair ( state, EpsilonType ( ) ) );
			}

			if ( localTransition != automaton.getLocalTransitions ( ).end ( ) )
				state = localTransition->second;
			else
				res = false;
		}

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;

		if ( res == false )
			break;
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType, class EpsilonType, class StateType >
std::tuple < bool, StateType, std::set < unsigned >, std::deque < SymbolType > > Run::calculateState ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	StateType state = automaton.getInitialState ( );
	std::deque < SymbolType > pushdownStore {
		automaton.getInitialSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose ) {
		std::clog << "State : " << state << std::endl;
		std::clog << "PushdownStore : " << pushdownStore << std::endl;
	}

	for ( auto symbolIter = string.getContent ( ).begin ( ); symbolIter != string.getContent ( ).end ( ); ) {
		auto transitions = automaton.getTransitionsFromState ( state );
		auto transition	 = transitions.begin ( );

		for ( ; transition != transitions.end ( ); transition++ ) {
			if ( ( std::get < 1 > ( transition->first ) != * symbolIter ) && !std::get < 1 > ( transition->first ).template is < EpsilonType > ( ) ) continue;

			if ( canPop ( pushdownStore, std::get < 2 > ( transition->first ) ) ) break;
		}

		if ( transition == transitions.end ( ) ) {
			res = false;
			break;
		}

		if ( common::GlobalData::verbose )
			std::clog << "Transition: " << * transition << std::endl;

		for ( unsigned j = 0; j < std::get < 2 > ( transition->first ).size ( ); j++ ) pushdownStore.pop_back ( );

		for ( const auto & symbol : std::make_reverse ( transition->second.second ) ) pushdownStore.push_back ( symbol );

		state = transition->second.first;

		if ( !std::get < 1 > ( transition->first ).template is < EpsilonType > ( ) ) {
			i++;
			symbolIter++;
		}

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose ) {
			std::clog << "State: " << state << std::endl;
			std::clog << "PushdownStore: " << pushdownStore << std::endl;
		}
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class SymbolType >
struct graphStructuredStack {
	graphStructuredStack ( std::shared_ptr < graphStructuredStack < SymbolType > > parent, SymbolType data ) : m_parent(parent), m_data(data) {}
	std::shared_ptr < graphStructuredStack < SymbolType > > m_parent;
	SymbolType m_data;
};

template < class SymbolType, class EpsilonType, class StateType >
std::tuple < bool, std::set < StateType >, std::set < std::vector < SymbolType > > > Run::calculateStates ( const automaton::NPDTA < SymbolType, EpsilonType, StateType > & automaton, const string::LinearString < SymbolType > & string ) {
	bool res = false;
	unsigned i = 0;
	std::set < unsigned > occurrences;
	StateType state = automaton . getInitialState ();
	typename std::vector<SymbolType>::const_iterator symbolIter = string . getContent () . begin ();
	std::shared_ptr < graphStructuredStack < SymbolType > > stackNode = std::make_shared < graphStructuredStack < SymbolType > > ( nullptr, * symbolIter );
	std::shared_ptr < graphStructuredStack < SymbolType > > outputNode = std::make_shared < graphStructuredStack < SymbolType > > ( nullptr, * symbolIter );

	if ( automaton . getFinalStates () . count ( state ) ) {
		occurrences . insert ( i );
	}

	std::deque < std::tuple < StateType, typename std::vector<SymbolType>::const_iterator, std::shared_ptr < graphStructuredStack < SymbolType > >, std::shared_ptr < graphStructuredStack < SymbolType > > > > bftQueue;
	auto configuration = std::make_tuple ( state, symbolIter, stackNode, outputNode );
	bftQueue . push_back ( configuration );

	std::set < std::vector < SymbolType > > allOutputs;
	std::set < StateType > states;

	while ( ! bftQueue . empty () ) {
		configuration = bftQueue . front (); bftQueue . pop_front ();
		state = std::get<0>(configuration);
		symbolIter = std::get<1>(configuration);
		stackNode = std::get<2>(configuration);
		outputNode = std::get<3>(configuration);

		if ( symbolIter == string . getContent () . end () ) {
			states . insert ( state );
			if ( automaton . getFinalStates () . count ( state ) && stackNode -> m_parent == NULL ) {
				res = true;
				std::vector < SymbolType > output;
				while ( outputNode->m_parent != NULL ) {
					output . push_back ( outputNode->m_data );
					outputNode = outputNode -> m_parent;
				}
				std::reverse ( output.begin(), output.end() );
				allOutputs . insert ( output );
			}
			continue;
		}

		for ( const auto & transition : automaton . getTransitionsFromState ( state ) ) {
			if ( std::get<1>(transition . first) != * symbolIter && ! std::get<1>(transition . first) . template is < EpsilonType >() )
				continue;

			const auto & pop = std::get<2>(transition . first);
			std::shared_ptr < graphStructuredStack < SymbolType > > stackNodeCopy = stackNode;
			std::shared_ptr < graphStructuredStack < SymbolType > > outputNodeCopy = outputNode;

			unsigned j = 0;
			for ( ; j < pop . size (); j ++ ) {
				if ( stackNodeCopy->m_parent == NULL || stackNodeCopy->m_data != pop . at (j) ) {
					break;
				}
				stackNodeCopy = stackNodeCopy -> m_parent;
			}
			if ( j == pop . size () ) {
				for ( const auto & elem : std::make_reverse ( std::get<1>(transition . second ) ) ) {
					stackNodeCopy = std::make_shared < graphStructuredStack < SymbolType > > ( stackNodeCopy, elem );
				}
				for ( const auto & elem : std::get<2>(transition . second) ) {
					outputNodeCopy = std::make_shared < graphStructuredStack < SymbolType > > ( outputNodeCopy, elem );
				}

				if ( ! std::get<1>(transition . first) . template is< EpsilonType > ( ) ) {
					configuration = std::make_tuple ( std::get<0>(transition . second), symbolIter + 1, stackNodeCopy, outputNodeCopy );
				} else {
					configuration = std::make_tuple ( std::get<0>(transition . second), symbolIter, stackNodeCopy, outputNodeCopy );
				}
				bftQueue . push_back ( configuration );
			} else {
				states . insert ( state );
			}
		}
	}

	return std::make_tuple ( res, states, allOutputs );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

} /* namespace run */

} /* namespace automaton */

#endif /* _AUTOMATON_RUN_H__ */
