/*
 * AutomataUnionCartesianProduct.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataUnionCartesianProduct.h"
#include <exception/CommonException.h>
#include <label/LabelPairLabel.h>

namespace automaton {

namespace transform {

automaton::Automaton AutomataUnionCartesianProduct::unification(const automaton::Automaton& first, const automaton::Automaton& second) {
	return dispatch(first.getData(), second.getData());
}

automaton::DFA<> AutomataUnionCartesianProduct::unification(const automaton::DFA<>& first, const automaton::DFA<>& second) {
	if(!first.isTotal() || !second.isTotal())
		throw exception::CommonException("Automata must be total to unify with cartesian product");

	label::Label q0(first.getInitialState(), second.getInitialState());
	automaton::DFA<> res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(label::Label(p, q));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getStates())
			res.addFinalState(label::Label(p, q));

	for(const auto& p : first.getStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(label::Label(p, q));

	for(const auto& state : res.getStates()) {
		const label::Label& label_p = static_cast<const label::LabelPairLabel&>(state.getData()).getData().first;
		const label::Label& label_q = static_cast<const label::LabelPairLabel&>(state.getData()).getData().second;

		for(const auto& tp : first.getTransitionsFromState(label::Label(label_p)))
			for(const auto& tq : second.getTransitionsFromState(label::Label(label_q)))
				if(tp.first.second == tq.first.second)
					res.addTransition(state, tp.first.second, label::Label(tp.second, tq.second));
	}

	return res;
}

auto AutomataUnionCartesianProductDFA = AutomataUnionCartesianProduct::RegistratorWrapper<automaton::DFA<>, automaton::DFA<>>(AutomataUnionCartesianProduct::unification);

automaton::NFA < > AutomataUnionCartesianProduct::unification(const automaton::NFA < > & first, const automaton::NFA < > & second) {
	if(!first.isTotal() || !second.isTotal())
		throw exception::CommonException("Automata must be total to unify with cartesian product");

	label::Label q0(first.getInitialState(), second.getInitialState());
	automaton::NFA < > res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(label::Label(p, q));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getStates())
			res.addFinalState(label::Label(p, q));

	for(const auto& p : first.getStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(label::Label(p, q));

	for(const auto& state : res.getStates()) {
		const label::Label& label_p = static_cast<const label::LabelPairLabel&>(state.getData()).getData().first;
		const label::Label& label_q = static_cast<const label::LabelPairLabel&>(state.getData()).getData().second;

		for(const auto& tp : first.getTransitionsFromState(label::Label(label_p)))
			for(const auto& tq : second.getTransitionsFromState(label::Label(label_q)))
				if(tp.first.second == tq.first.second)
					for(const auto& p : tp.second)
						for(const auto& q : tq.second)
							res.addTransition(state, tp.first.second, label::Label(p, q));
	}

	return res;
}

auto AutomataUnionCartesianProductNFA = AutomataUnionCartesianProduct::RegistratorWrapper<automaton::NFA < > , automaton::NFA < > >(AutomataUnionCartesianProduct::unification);

} /* namespace transform */

} /* namespace automaton */
