/*
 * RHPDAToPDA.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Jan Travnicek
 */

#include "RHPDAToPDA.h"

#include <exception/CommonException.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/NPDA.h>

#include <alphabet/BottomOfTheStackSymbol.h>

#include <set>
#include <map>
#include <queue>
#include <iostream>

namespace automaton {

template < class T >
void constructTransitions ( const std::tuple < DefaultStateType, DefaultSymbolType, std::vector < DefaultSymbolType > > & stFirst, const std::map < DefaultStateType, std::set < std::tuple < std::vector < DefaultSymbolType >, DefaultStateType, std::vector < DefaultSymbolType > > > > & epsilonTransitions, const DefaultStateType & toState, std::vector < DefaultSymbolType > pops, std::vector < DefaultSymbolType > pushes, T & res ) {
	auto epsIter = epsilonTransitions.find ( toState );

	if ( ( epsIter != epsilonTransitions.end ( ) ) && !epsIter->second.empty ( ) ) {
		if ( epsIter->second.size ( ) == 1 ) {
			const auto & epsilonT = * epsIter->second.begin ( );

			pops.insert ( pops.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
			pushes.insert ( pushes.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

			constructTransitions ( stFirst, epsilonTransitions, std::get < 1 > ( epsilonT ), std::move ( pops ), std::move ( pushes ), res );
		} else {
			for ( const auto & epsilonT : epsIter->second ) {
				std::vector < DefaultSymbolType > popsCopy = pops;
				std::vector < DefaultSymbolType > pushesCopy = pushes;

				popsCopy.insert ( popsCopy.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
				pushesCopy.insert ( pushesCopy.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

				constructTransitions ( stFirst, epsilonTransitions, std::get < 1 > ( epsilonT ), std::move ( popsCopy ), std::move ( pushesCopy ), res );
			}
		}
	} else {
		res.addState ( std::get < 0 > ( stFirst ) );
		res.addState ( toState );

		res.addTransition ( std::get < 0 > ( stFirst ), std::get < 1 > ( stFirst ), std::move ( pops ), toState, std::move ( pushes ) );
	}
}

automaton::DPDA < > RHPDAToPDA::convert ( const automaton::RealTimeHeightDeterministicDPDA < > & pda ) {
	std::map < std::tuple < DefaultStateType, DefaultSymbolType, std::vector < DefaultSymbolType > >, std::set < std::pair < DefaultStateType, std::vector < DefaultSymbolType > > > > readingTransitions;
	std::map < DefaultStateType, std::set < std::tuple < std::vector < DefaultSymbolType >, DefaultStateType, std::vector < DefaultSymbolType > > > > epsilonTransitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < DefaultEpsilonType > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];
			const auto & to = transition.second;
			epsT.insert ( std::make_tuple ( std::vector < DefaultSymbolType > { }, to.first, std::vector < DefaultSymbolType > { to.second } ) );
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < DefaultSymbolType > ( ), std::vector < DefaultSymbolType > { } )];
			const auto & to = transition.second;
			readT.insert ( std::make_pair ( to.first, std::vector < DefaultSymbolType > { to.second } ) );
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < DefaultEpsilonType > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];
			const auto & to = transition.second;
			epsT.insert ( std::make_tuple ( std::vector < DefaultSymbolType > { }, to, std::vector < DefaultSymbolType > { } ) );
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < DefaultSymbolType > ( ), std::vector < DefaultSymbolType > { } )];
			const auto & to = transition.second;
			readT.insert ( std::make_pair ( to, std::vector < DefaultSymbolType > { } ) );
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		if ( std::get < 2 > ( transition.first ) == alphabet::BottomOfTheStackSymbol::instance < DefaultSymbolType > ( ) ) continue;

		if ( std::get < 1 > ( transition.first ).is < DefaultEpsilonType > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];
			const auto & to = transition.second;
			epsT.insert ( std::make_tuple ( std::vector < DefaultSymbolType > { std::get < 2 > ( transition.first ) }, to, std::vector < DefaultSymbolType > { } ) );
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < DefaultSymbolType > ( ), std::vector < DefaultSymbolType > { std::get < 2 > ( transition.first ) } )];
			const auto & to = transition.second;
			readT.insert ( std::make_pair ( to, std::vector < DefaultSymbolType > { } ) );
		}
	}

	for ( const auto & st : epsilonTransitions ) {
		bool allPops = true;

		for ( const auto & elems : st.second )
			allPops &= std::get < 2 > ( elems ).empty ( ); // if some pushes it will clear allPops variable

		if ( st.second.size ( ) == 1 ) continue;

		if ( ( st.second.size ( ) > 1 ) && allPops ) continue;

		throw exception::CommonException ( "Temporary states has more than one leaving transition" );
	}

	if ( epsilonTransitions[pda.getInitialState ( )].empty ( ) )
		throw exception::CommonException ( "Cannot determine initial pushdown store symbol" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	const auto & st = * epsilonTransitions[pda.getInitialState ( )].begin ( );

	std::vector < DefaultSymbolType > pops ( std::get < 0 > ( st ).begin ( ), std::get < 0 > ( st ).end ( ) );
	std::vector < DefaultSymbolType > pushes ( std::get < 2 > ( st ).rbegin ( ), std::get < 2 > ( st ).rend ( ) );

	DefaultStateType toState = std::get < 1 > ( st );

	while ( !epsilonTransitions[toState].empty ( ) ) {
		const auto & epsilonT = * epsilonTransitions[toState].begin ( );

		pops.insert ( pops.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
		pushes.insert ( pushes.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

		toState = std::get < 1 > ( epsilonT );
	}

	if ( ( pops.size ( ) != 0 ) && ( pushes.size ( ) != 1 ) )
		throw exception::CommonException ( "Cannot convert" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	automaton::DPDA < > res ( toState, pushes[0] );

	res.setInputAlphabet ( pda.getInputAlphabet ( ) );
	res.setPushdownStoreAlphabet ( pda.getPushdownStoreAlphabet ( ) );
	res.removePushdownStoreSymbol ( pda.getBottomOfTheStackSymbol ( ) );

	for ( const auto & transition : readingTransitions )
		for ( const auto & to : transition.second ) {
			pops = std::vector < DefaultSymbolType > ( std::get < 2 > ( transition.first ).begin ( ), std::get < 2 > ( transition.first ).end ( ) );
			pushes = std::vector < DefaultSymbolType > ( to.second.rbegin ( ), to.second.rend ( ) );

			constructTransitions ( transition.first, epsilonTransitions, to.first, std::move ( pops ), std::move ( pushes ), res );
		}

	res.setFinalStates ( pda.getFinalStates ( ) );

	return res;
}

auto RHPDAToPDARealTimeHeightDeterministicDPDA = RHPDAToPDA::RegistratorWrapper < automaton::DPDA < >, automaton::RealTimeHeightDeterministicDPDA < > > ( RHPDAToPDA::convert );

 // This may not work correctly -- generation of initial state and initial symbol
automaton::NPDA < > RHPDAToPDA::convert ( const automaton::RealTimeHeightDeterministicNPDA < > & pda ) {
	std::map < std::tuple < DefaultStateType, DefaultSymbolType, std::vector < DefaultSymbolType > >, std::set < std::pair < DefaultStateType, std::vector < DefaultSymbolType > > > > readingTransitions;
	std::map < DefaultStateType, std::set < std::tuple < std::vector < DefaultSymbolType >, DefaultStateType, std::vector < DefaultSymbolType > > > > epsilonTransitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < DefaultEpsilonType > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];

			for ( const auto & to : transition.second ) {
				epsT.insert ( std::make_tuple ( std::vector < DefaultSymbolType > { }, to.first, std::vector < DefaultSymbolType > { to.second } ) );
			}
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < DefaultSymbolType > ( ), std::vector < DefaultSymbolType > { } )];

			for ( const auto & to : transition.second ) {
				readT.insert ( std::make_pair ( to.first, std::vector < DefaultSymbolType > { to.second } ) );
			}
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		if ( std::get < 1 > ( transition.first ).is < DefaultEpsilonType > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];

			for ( const auto & to : transition.second ) {
				epsT.insert ( std::make_tuple ( std::vector < DefaultSymbolType > { }, to, std::vector < DefaultSymbolType > { } ) );
			}
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < DefaultSymbolType > ( ), std::vector < DefaultSymbolType > { } )];

			for ( const auto & to : transition.second ) {
				readT.insert ( std::make_pair ( to, std::vector < DefaultSymbolType > { } ) );
			}
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		if ( std::get < 2 > ( transition.first ) == alphabet::BottomOfTheStackSymbol::instance < DefaultSymbolType > ( ) ) continue;

		if ( std::get < 1 > ( transition.first ).is < DefaultEpsilonType > ( ) ) {
			auto & epsT = epsilonTransitions[std::get < 0 > ( transition.first )];

			for ( const auto & to : transition.second ) {
				epsT.insert ( std::make_tuple ( std::vector < DefaultSymbolType > { std::get < 2 > ( transition.first ) }, to, std::vector < DefaultSymbolType > { } ) );
			}
		} else {
			auto & readT = readingTransitions[std::make_tuple ( std::get < 0 > ( transition.first ), std::get < 1 > ( transition.first ).get < DefaultSymbolType > ( ), std::vector < DefaultSymbolType > { std::get < 2 > ( transition.first ) } )];

			for ( const auto & to : transition.second ) {
				readT.insert ( std::make_pair ( to, std::vector < DefaultSymbolType > { } ) );
			}
		}
	}

	for ( const auto & st : epsilonTransitions ) {
		bool allPops = true;

		for ( const auto & elems : st.second )
			allPops &= std::get < 2 > ( elems ).empty ( ); // if some pushes it will clear allPops variable

		if ( ( st.second.size ( ) != 1 ) && !allPops ) throw exception::CommonException ( "Temporary states has more than one leaving transition" );
	}

	if ( pda.getInitialStates ( ).size ( ) != 1 )
		throw exception::CommonException ( "Cannot convert" );

	const DefaultStateType & initialState = * pda.getInitialStates ( ).begin ( );

	if ( epsilonTransitions[initialState].empty ( ) )
		throw exception::CommonException ( "Cannot convert" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	const auto & st = * epsilonTransitions[initialState].begin ( );

	std::vector < DefaultSymbolType > pops ( std::get < 0 > ( st ).begin ( ), std::get < 0 > ( st ).end ( ) );
	std::vector < DefaultSymbolType > pushes ( std::get < 2 > ( st ).rbegin ( ), std::get < 2 > ( st ).rend ( ) );

	DefaultStateType toState = std::get < 1 > ( st );

	while ( !epsilonTransitions[toState].empty ( ) ) {
		const auto & epsilonT = * epsilonTransitions[toState].begin ( );

		pops.insert ( pops.end ( ), std::get < 0 > ( epsilonT ).begin ( ), std::get < 0 > ( epsilonT ).end ( ) );
		pushes.insert ( pushes.begin ( ), std::get < 2 > ( epsilonT ).rbegin ( ), std::get < 2 > ( epsilonT ).rend ( ) );

		toState = std::get < 1 > ( epsilonT );
	}

	if ( ( pops.size ( ) != 0 ) && ( pushes.size ( ) != 1 ) )
		throw exception::CommonException ( "Cannot convert" );

	// -------------------------------------------------------------------- initial state and initial stack symbol

	automaton::NPDA < > res ( toState, pushes[0] );

	res.setInputAlphabet ( pda.getInputAlphabet ( ) );
	res.setPushdownStoreAlphabet ( pda.getPushdownStoreAlphabet ( ) );
	res.removePushdownStoreSymbol ( pda.getBottomOfTheStackSymbol ( ) );

	for ( const auto & transition : readingTransitions )
		for ( const auto & to : transition.second ) {
			pops = std::vector < DefaultSymbolType > ( std::get < 2 > ( transition.first ).begin ( ), std::get < 2 > ( transition.first ).end ( ) );
			pushes = std::vector < DefaultSymbolType > ( to.second.rbegin ( ), to.second.rend ( ) );

			constructTransitions ( transition.first, epsilonTransitions, to.first, std::move ( pops ), std::move ( pushes ), res );
		}

	res.setFinalStates ( pda.getFinalStates ( ) );

	return res;
}

auto RHPDAToPDARealTimeHeightDeterministicNPDA = RHPDAToPDA::RegistratorWrapper < automaton::NPDA < >, automaton::RealTimeHeightDeterministicNPDA < > > ( RHPDAToPDA::convert );

automaton::Automaton RHPDAToPDA::convert ( const Automaton & automaton ) {
	return dispatch ( automaton.getData ( ) );
}

} /* namespace automaton */

namespace alib {

auto DPDAFromRealTimeHeightDeterministicNPDA = castApi::CastRegister < automaton::DPDA < >, automaton::RealTimeHeightDeterministicDPDA < > > ( automaton::RHPDAToPDA::convert );
auto NPDAFromRealTimeHeightDeterministicNPDA = castApi::CastRegister < automaton::NPDA < >, automaton::RealTimeHeightDeterministicNPDA < > > ( automaton::RHPDAToPDA::convert );

} /* namespace alib */
