/*
 * UnproductiveSymbolsRemover.h
 *
 *  Created on: 22. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef UNPRODUCTIVE_SYMBOLS_REMOVER_H_
#define UNPRODUCTIVE_SYMBOLS_REMOVER_H_

#include <core/multipleDispatch.hpp>
#include <grammar/Grammar.h>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <set>
#include <algorithm>

#include <grammar/properties/ProductiveNonterminals.h>
#include <exception/CommonException.h>

namespace grammar {

namespace simplify {

/**
 * Implements algorithms from Melichar, chapter 3.3
 */
class UnproductiveSymbolsRemover : public std::SingleDispatch<UnproductiveSymbolsRemover, grammar::Grammar, const grammar::GrammarBase &> {
public:
	static grammar::Grammar remove( const grammar::Grammar & automaton );

	/**
	 * Removes unproductive (or useless - terminology) symbols - Melichar 3.12
	 */
	template < class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static T remove( const T & grammar );
};

template < class T, class SymbolType >
T UnproductiveSymbolsRemover::remove( const T & grammar ) {
	// 1.
	std::set<SymbolType> Nt = grammar::properties::ProductiveNonterminals::getProductiveNonterminals( grammar );

	T ret(grammar.getInitialSymbol( ) );

	for( const auto & symbol : Nt )
		ret.addNonterminalSymbol( symbol );

	for( const auto & symbol : grammar.getTerminalAlphabet( ) )
		ret.addTerminalSymbol( symbol );

	const std::set<SymbolType> & terminals = ret.getTerminalAlphabet( );
	for( const auto & rule : grammar.getRawRules( ) ) {
		if( Nt.count( rule.first ) ) {
			for( const auto & rhs : rule.second ) {
				if( all_of( rhs.begin( ), rhs.end( ), [ Nt, terminals ]( const SymbolType & symbol ) {
						return Nt.count( symbol ) || terminals.count( symbol );
				} ) )
					ret.addRawRule( rule.first, rhs );
			}
		}
	}


	/* if( ! G1.getNonTerminalSymbols( ) . count( grammar.getInitialSymbol( ) ) )
		throw CommonException( "Starting symbol of grammar was marked as unproductive and therefore it was removed." ); */

	// 2.
	return ret;
}

} /* namespace simplify */

} /* namespace grammar */

#endif /* UNPRODUCTIVE_SYMBOLS_REMOVER_H_ */
