/*
 * ToAutomatonGlushkov.cpp
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#include "ToAutomatonGlushkov.h"

#include "label/Label.h"
#include "label/LabelPairLabel.h"
#include <automaton/Automaton.h>

#include <global/GlobalData.h>

#include "../glushkov/GlushkovFirst.h"
#include "../glushkov/GlushkovFollow.h"
#include "../glushkov/GlushkovIndexate.h"
#include "../glushkov/GlushkovLast.h"

#include "../properties/RegExpEpsilon.h"
#include <exception/CommonException.h>

namespace regexp {

namespace convert {

automaton::NFA < > ToAutomatonGlushkov::convert ( const regexp::RegExp & regexp ) {
	return dispatch ( regexp.getData ( ) );
}

automaton::NFA < > ToAutomatonGlushkov::convert ( const regexp::UnboundedRegExp < > & regexp ) {
	label::Label q0 ( label::LabelPairLabel ( std::make_pair ( label::Label ( 'q' ), label::Label ( 0 ) ) ) );
	automaton::NFA < > automaton ( q0 );

	 // step 1
	automaton.setInputAlphabet ( regexp.getAlphabet ( ) );

	regexp::UnboundedRegExp < > indexedRegExp = regexp::GlushkovIndexate::index ( regexp );

	 // steps 2, 3, 4
	const std::set < regexp::UnboundedRegExpSymbol < alphabet::Symbol > > first = regexp::GlushkovFirst::first ( indexedRegExp );
	const std::set < regexp::UnboundedRegExpSymbol < alphabet::Symbol > > last = regexp::GlushkovLast::last ( indexedRegExp );

	// \e in q0 check is in step 7

	 // step 5
	for ( auto const & symbol : indexedRegExp.getAlphabet ( ) )
		automaton.addState ( label::Label ( symbol ) );

	 // step 6
	for ( auto const & symbol : first )
		automaton.addTransition ( q0, regexp::GlushkovIndexate::getSymbolFromGlushkovPair ( symbol.getSymbol ( ) ), label::Label ( symbol.getSymbol ( ) ) );

	for ( const auto & x : indexedRegExp.getAlphabet ( ) )
		for ( const auto & f : regexp::GlushkovFollow::follow ( indexedRegExp, UnboundedRegExpSymbol < alphabet::Symbol > ( x ) ) ) {
			const alphabet::Symbol & p = x;
			const alphabet::Symbol & q = f.getSymbol ( );

			automaton.addTransition ( label::Label ( p ), regexp::GlushkovIndexate::getSymbolFromGlushkovPair ( q ), label::Label ( q ) );
		}

	// step 7

	for ( auto const & symbol : last )
		automaton.addFinalState ( label::Label ( symbol.getSymbol ( ) ) );

	if ( regexp::properties::RegExpEpsilon::languageContainsEpsilon ( regexp ) )
		automaton.addFinalState ( q0 );

	if ( common::GlobalData::verbose ) {
		std::clog << "First:" << first << std::endl;
		std::clog << "Last: " << last << std::endl;

		if ( regexp::properties::RegExpEpsilon::languageContainsEpsilon ( regexp ) )
			std::clog << "      q0 because #E in L(RE)" << std::endl;

		for ( const auto & x : indexedRegExp.getAlphabet ( ) )
			std::clog << "Follow(" << x << ") = " << regexp::GlushkovFollow::follow ( indexedRegExp, UnboundedRegExpSymbol < alphabet::Symbol > ( x ) ) << std::endl;
	}

	return automaton;
}

auto ToAutomatonGlushkovUnboundedRegExp = ToAutomatonGlushkov::RegistratorWrapper < automaton::NFA < >, regexp::UnboundedRegExp < > > ( ToAutomatonGlushkov::convert );

automaton::NFA < > ToAutomatonGlushkov::convert ( const regexp::FormalRegExp < > & /* regexp */ ) {
	throw exception::CommonException ( "Glushkov: Converting FormalRegExp NYI" ); // TODO
}

auto ToAutomatonGlushkovFormalRegExp = ToAutomatonGlushkov::RegistratorWrapper < automaton::NFA < >, regexp::FormalRegExp < > > ( ToAutomatonGlushkov::convert );

} /* namespace convert */

} /* namespace regexp */
