#include <list>
#include "determinizeTest.h"

#include <automaton/FSM/NFA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <alphabet/BottomOfTheStackSymbol.h>

#include "automaton/determinize/Determinize.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( determinizeTest, "automaton" );
CPPUNIT_TEST_SUITE_REGISTRATION( determinizeTest );

void determinizeTest::setUp() {
}

void determinizeTest::tearDown() {
}

void determinizeTest::testDeterminizeNFA() {
  automaton::NFA < > automaton(label::Label(1));

  automaton.addState(label::Label(1));
  automaton.addState(label::Label(2));
  automaton.addState(label::Label(3));
  automaton.addInputSymbol(alphabet::Symbol("a"));
  automaton.addInputSymbol(alphabet::Symbol("b"));

  automaton.addTransition(label::Label(1), alphabet::Symbol("a"), label::Label(2));
  automaton.addTransition(label::Label(2), alphabet::Symbol("b"), label::Label(1));

  automaton.addFinalState(label::Label(3));

  automaton::DFA<> determinized = automaton::determinize::Determinize::determinize(automaton);

  CPPUNIT_ASSERT(determinized.getStates().size() == 3);

}

void determinizeTest::testDeterminizeIDPDA() {
  automaton::InputDrivenNPDA < > automaton(label::Label(1), alphabet::Symbol('S'));

  automaton.addInputSymbol(alphabet::Symbol("a"));
  automaton.addInputSymbol(alphabet::Symbol("b"));

  automaton.setPushdownStoreOperation(alphabet::Symbol("a"), std::vector<alphabet::Symbol>{}, std::vector<alphabet::Symbol> {});
  automaton.setPushdownStoreOperation(alphabet::Symbol("b"), std::vector<alphabet::Symbol>{}, std::vector<alphabet::Symbol> {});

  automaton.addState(label::Label(1));
  automaton.addState(label::Label(2));
  automaton.addState(label::Label(3));
  automaton.addTransition(label::Label(1), alphabet::Symbol("a"), label::Label(2));
  automaton.addTransition(label::Label(2), alphabet::Symbol("b"), label::Label(1));

  automaton.addFinalState(label::Label(3));

  automaton::InputDrivenDPDA < > determinized = automaton::determinize::Determinize::determinize(automaton);

  CPPUNIT_ASSERT(determinized.getStates().size() == 3);
}

void determinizeTest::testDeterminizeVPA() {
  automaton::VisiblyPushdownNPDA < > automaton(alphabet::Symbol(alphabet::BottomOfTheStackSymbol {}));

  automaton.addCallInputSymbol(alphabet::Symbol('a'));
  automaton.addReturnInputSymbol(alphabet::Symbol('^'));

  automaton.addPushdownStoreSymbol(alphabet::Symbol('A'));
  automaton.addPushdownStoreSymbol(alphabet::Symbol('B'));
  automaton.addPushdownStoreSymbol(alphabet::Symbol('C'));
  automaton.addPushdownStoreSymbol(alphabet::Symbol('D'));
  automaton.addPushdownStoreSymbol(alphabet::Symbol('E'));
  automaton.addPushdownStoreSymbol(alphabet::Symbol('F'));
  automaton.addPushdownStoreSymbol(alphabet::Symbol('T'));

  automaton.addState(label::Label(0));
  automaton.addState(label::Label(1));
  automaton.addState(label::Label(2));
  automaton.addState(label::Label(3));
  automaton.addState(label::Label(4));
  automaton.addState(label::Label(5));
  automaton.addState(label::Label(6));


  automaton.addCallTransition(label::Label(0), alphabet::Symbol('a'), label::Label(0), alphabet::Symbol('A'));
  automaton.addReturnTransition(label::Label(0), alphabet::Symbol('^'), alphabet::Symbol('A'), label::Label(0));

  automaton.addCallTransition(label::Label(0), alphabet::Symbol('a'), label::Label(1), alphabet::Symbol('B'));

  automaton.addCallTransition(label::Label(1), alphabet::Symbol('a'), label::Label(2), alphabet::Symbol('C'));
  automaton.addCallTransition(label::Label(2), alphabet::Symbol('a'), label::Label(2), alphabet::Symbol('D'));
  automaton.addReturnTransition(label::Label(2), alphabet::Symbol('^'), alphabet::Symbol('D'), label::Label(2));
  automaton.addReturnTransition(label::Label(2), alphabet::Symbol('^'), alphabet::Symbol('C'), label::Label(3));

  automaton.addCallTransition(label::Label(3), alphabet::Symbol('a'), label::Label(4), alphabet::Symbol('E'));
  automaton.addCallTransition(label::Label(4), alphabet::Symbol('a'), label::Label(4), alphabet::Symbol('F'));
  automaton.addReturnTransition(label::Label(4), alphabet::Symbol('^'), alphabet::Symbol('F'), label::Label(4));
  automaton.addReturnTransition(label::Label(4), alphabet::Symbol('^'), alphabet::Symbol('E'), label::Label(5));

  automaton.addReturnTransition(label::Label(5), alphabet::Symbol('^'), alphabet::Symbol('B'), label::Label(6));

  automaton.addInitialState(label::Label(0));
  automaton.addFinalState(label::Label(4));
}

void determinizeTest::testDeterminizeNFTA() {
  automaton::NFTA < > automaton;

  const std::ranked_symbol < > a ('a', 2);
  const std::ranked_symbol < > b ('b', 1);
  const std::ranked_symbol < > c ('c', 0);
  const std::set<std::ranked_symbol < > > alphabet {a, b, c};
  automaton.setInputAlphabet(alphabet);

  automaton.addState(label::Label(1));
  automaton.addState(label::Label(2));
  automaton.addState(label::Label(3));

  std::vector<label::Label> a1States = {label::Label(1), label::Label(3)};
  automaton.addTransition(a, a1States, label::Label(1));
  automaton.addTransition(a, a1States, label::Label(3));
  std::vector<label::Label> a2States = {label::Label(3), label::Label(3)};
  automaton.addTransition(a, a2States, label::Label(2));
  std::vector<label::Label> bStates = {label::Label(2)};
  automaton.addTransition(b, bStates, label::Label(1));
  std::vector<label::Label> cStates;
  automaton.addTransition(c, cStates, label::Label(3));

  automaton.addFinalState(label::Label(3));


  automaton::DFTA < > determinized = automaton::determinize::Determinize::determinize(automaton);

  CPPUNIT_ASSERT(determinized.getStates().size() == 5);
  CPPUNIT_ASSERT(determinized.getFinalStates().size() == 3);
  CPPUNIT_ASSERT(determinized.getTransitions().size() == 15);
}
