#include "GrammarPropertiesTest.h"

#include <label/PrimitiveLabel.h>
#include <grammar/Grammar.h>
#include <alphabet/LabeledSymbol.h>
#include <alphabet/Symbol.h>
#include <grammar/ContextFree/CFG.h>
#include <grammar/Regular/LeftLG.h>

#include "grammar/properties/NullableNonterminals.h"
#include "grammar/properties/NonterminalUnitRuleCycle.h"
#include "grammar/properties/RecursiveNonterminal.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( GrammarPropertiesTest, "grammar" );
CPPUNIT_TEST_SUITE_REGISTRATION( GrammarPropertiesTest );

void GrammarPropertiesTest::setUp() {
}

void GrammarPropertiesTest::tearDown() {
}

void GrammarPropertiesTest::testNullable() {
	DefaultSymbolType X = DefaultSymbolType('X');
	DefaultSymbolType Y = DefaultSymbolType('Y');
	DefaultSymbolType Z = DefaultSymbolType('Z');
	DefaultSymbolType d = DefaultSymbolType('d');

	grammar::CFG < > grammar(X);
	grammar.setTerminalAlphabet({d});
	grammar.setNonterminalAlphabet({{X, Y, Z}});
	grammar.setInitialSymbol(X);

	grammar.addRule(X, std::vector<DefaultSymbolType>{ d });
	grammar.addRule(X, std::vector<DefaultSymbolType>{ Y });
	grammar.addRule(Y, std::vector<DefaultSymbolType>{ d });
	grammar.addRule(Y, std::vector<DefaultSymbolType>{ });
	grammar.addRule(Z, std::vector<DefaultSymbolType>{ d });
	grammar.addRule(Z, std::vector<DefaultSymbolType>{ X, Y, Z });

	std::set<DefaultSymbolType> res = {X, Y};
	CPPUNIT_ASSERT(res == grammar::properties::NullableNonterminals::getNullableNonterminals(grammar));
}

void GrammarPropertiesTest::testUnitRules() {
	DefaultSymbolType S = DefaultSymbolType('S');
	DefaultSymbolType A = DefaultSymbolType('A');
	DefaultSymbolType B = DefaultSymbolType('B');
	DefaultSymbolType C = DefaultSymbolType('C');
	DefaultSymbolType a = DefaultSymbolType('a');
	DefaultSymbolType b = DefaultSymbolType('b');

	grammar::LeftLG < > llg(S);
	llg.setTerminalAlphabet({a, b});
	llg.setNonterminalAlphabet({S, A, B, C});
	llg.setInitialSymbol(S);

	llg.addRule(S, std::make_pair(A, std::vector<DefaultSymbolType>{}));
	llg.addRule(S, std::make_pair(B, std::vector<DefaultSymbolType>{}));
	llg.addRule(A, std::make_pair(A, std::vector<DefaultSymbolType>{a}));
	llg.addRule(A, std::make_pair(B, std::vector<DefaultSymbolType>{}));
	llg.addRule(A, std::make_pair(C, std::vector<DefaultSymbolType>{}));
	llg.addRule(B, std::make_pair(B, std::vector<DefaultSymbolType>{b, a}));
	llg.addRule(B, std::make_pair(C, std::vector<DefaultSymbolType>{b}));
	llg.addRule(C, {b});
	llg.addRule(C, std::make_pair(C, std::vector<DefaultSymbolType>{a}));

	std::set<DefaultSymbolType> N_S = {S, A, B, C};
	std::set<DefaultSymbolType> N_A = {A, B, C};
	std::set<DefaultSymbolType> N_B = {B};
	std::set<DefaultSymbolType> N_C = {C};

	CPPUNIT_ASSERT(N_S == grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(llg, S));
	CPPUNIT_ASSERT(N_A == grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(llg, A));
	CPPUNIT_ASSERT(N_B == grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(llg, B));
	CPPUNIT_ASSERT(N_C == grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(llg, C));
}

void GrammarPropertiesTest::testRecursive() {
	{
		DefaultSymbolType A = DefaultSymbolType('A');
		DefaultSymbolType B = DefaultSymbolType('B');
		DefaultSymbolType C = DefaultSymbolType('C');
		DefaultSymbolType D = DefaultSymbolType('D');

		DefaultSymbolType a = DefaultSymbolType('a');
		DefaultSymbolType b = DefaultSymbolType('b');

		grammar::CFG < > cfg(A);
		cfg.setTerminalAlphabet({a, b});
		cfg.setNonterminalAlphabet({A, B, C, D});
		cfg.setInitialSymbol(A);

		cfg.addRule(A, {B});
		cfg.addRule(B, {C});
		cfg.addRule(C, {D});
		cfg.addRule(D, {A});
		cfg.addRule(D, {});

		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, A));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, B));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, C));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, D));
	}
	{
		DefaultSymbolType A = DefaultSymbolType('A');
		DefaultSymbolType B = DefaultSymbolType('B');
		DefaultSymbolType C = DefaultSymbolType('C');
		DefaultSymbolType D = DefaultSymbolType('D');

		DefaultSymbolType a = DefaultSymbolType('a');
		DefaultSymbolType b = DefaultSymbolType('b');

		grammar::CFG < > cfg(A);
		cfg.setTerminalAlphabet({a, b});
		cfg.setNonterminalAlphabet({A, B, C, D});
		cfg.setInitialSymbol(A);

		cfg.addRule(A, {B});
		cfg.addRule(B, {C});
		cfg.addRule(C, {D});
		cfg.addRule(D, {B});
		cfg.addRule(D, {});

		CPPUNIT_ASSERT(!grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, A));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, B));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, C));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, D));
	}
	{
		DefaultSymbolType A = DefaultSymbolType('A');
		DefaultSymbolType B = DefaultSymbolType('B');
		DefaultSymbolType C = DefaultSymbolType('C');
		DefaultSymbolType D = DefaultSymbolType('D');

		DefaultSymbolType a = DefaultSymbolType('a');
		DefaultSymbolType b = DefaultSymbolType('b');

		grammar::CFG < > cfg(A);
		cfg.setTerminalAlphabet({a, b});
		cfg.setNonterminalAlphabet({A, B, C, D});
		cfg.setInitialSymbol(A);

		cfg.addRule(A, {B});
		cfg.addRule(B, {C, D, A});
		cfg.addRule(C, {a});
		cfg.addRule(C, {});
		cfg.addRule(D, {b});
		cfg.addRule(D, {});

		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, A));
		CPPUNIT_ASSERT(grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, B));
		CPPUNIT_ASSERT(!grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, C));
		CPPUNIT_ASSERT(!grammar::properties::RecursiveNonterminal::isNonterminalRecursive(cfg, D));
	}
}
