/*
 * BellmanFord.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 */

#include "BellmanFord.h"

#include <exception/CommonException.h>

namespace graph
{

namespace shortestpath
{

class Data
{
public:
	Node start;
	BellmanFord::Result out;
};

typedef std::unordered_map<Node, std::unordered_map<Node, int>> weights_t;

static BellmanFord::Result bellmanford_impl(const DirectedGraph &graph, const Node &start)
{
	BellmanFord::Result d; // distances

	d[start] = 0;

	for (unsigned i = 0; i < graph.getNodes().size(); ++i) {
		for (const DirectedEdge &e : graph.getEdges()) {
			int ew = graph.getEdgeValue(e);
			const Node &u = e.getFromNode();
			const Node &v = e.getToNode();

			auto du = d.find(u);
			auto dv = d.find(v);

			if (du == d.end()) {
				continue;
			}

			if (dv == d.end() || du->second + ew < dv->second) {
				d[v] = du->second + ew;
			}
		}
	}

	for (const DirectedEdge &e : graph.getEdges()) {
		const Node &u = e.getFromNode();
		const Node &v = e.getToNode();

		if (d.at(u) + graph.getEdgeValue(e) < d.at(v)) {
			throw exception::CommonException("BellmanFord: Found negative-weight cycle!");
		}
	}

	return d;
}

static BellmanFord::Result bellmanford_impl(const UndirectedGraph &graph, const Node &start)
{
	BellmanFord::Result d; // distances

	d[start] = 0;

	for (unsigned i = 0; i < graph.getNodes().size(); ++i) {
		for (const UndirectedEdge &e : graph.getEdges()) {
			int ew = graph.getEdgeValue(e);
			const Node &u = e.getFirstNode();
			const Node &v = e.getSecondNode();

			auto f = [&ew, &d](const Node &up, const Node &vp) {
				auto du = d.find(up);
				auto dv = d.find(vp);

				if (du == d.end()) {
					return;
				}

				if (dv == d.end() || du->second + ew < dv->second) {
					d[vp] = du->second + ew;
				}
			};

			f(u, v);
			f(v, u);
		}
	}

	for (const UndirectedEdge &e : graph.getEdges()) {
		const Node &u = e.getFirstNode();
		const Node &v = e.getSecondNode();

		auto f = [&e, &d, &graph](const Node &up, const Node &vp) {
			if (d.at(up) + graph.getEdgeValue(e) < d.at(vp)) {
				throw exception::CommonException("BellmanFord: Found negative-weight cycle!");
			}
		};

		f(u, v);
		f(v, u);

	}

	return d;
}

BellmanFord::Result BellmanFord::bellmanford(const Graph &graph, const Node &start)
{
	return dispatch(graph.getData(), start);
}

BellmanFord::Result BellmanFord::bellmanford(const DirectedGraph &graph, const Node &start)
{
	return bellmanford_impl(graph, start);
}

auto BellmanFordDirectedGraph = BellmanFord::RegistratorWrapper<BellmanFord::Result, graph::DirectedGraph>(BellmanFord::bellmanford);

BellmanFord::Result BellmanFord::bellmanford(const UndirectedGraph &graph, const Node &start)
{
	return bellmanford_impl(graph, start);
}

auto BellmanFordUndirectedGraph = BellmanFord::RegistratorWrapper<BellmanFord::Result, graph::UndirectedGraph>(BellmanFord::bellmanford);

} // namespace shortestpath

} // namespace graph
