/*
 * Author: Radovan Cerveny
 */

#include "BackwardDAWGMatching.h"
#include "SuffixAutomaton.h"

#include <string/LinearString.h>
#include <alphabet/Symbol.h>

#include <algorithm>
#include <map>
#include <bitset>
#include <measure>

namespace stringology {

namespace exact {

std::set < unsigned > BackwardDAWGMatching::match ( const string::String & subject, const string::String & pattern ) {
    return dispatch ( subject.getData ( ), pattern.getData ( ) );
}

std::set < unsigned > BackwardDAWGMatching::match ( const string::LinearString < > & subject, const string::LinearString < > & pattern ) {
    std::set < unsigned > occ;

    measurements::start ( "Preprocess", measurements::Type::PREPROCESS );

    auto patternData = pattern.getContent ( );

    reverse ( patternData.begin ( ), patternData.end ( ) );

    const string::LinearString < > reversedPattern ( std::move ( patternData ) );

    automaton::DFA<> suffixAutomaton = SuffixAutomaton::construct ( reversedPattern );

    measurements::end ( );

    measurements::start ( "Algorithm", measurements::Type::ALGORITHM );

    const label::Label failState = label::Label ( -1 );

    size_t posInSubject = 0;

    while ( posInSubject <= subject.getContent ( ).size ( ) - pattern.getContent ( ).size ( ) ) {
        label::Label currentState = suffixAutomaton.getInitialState ( );

        size_t posInPattern = reversedPattern.getContent ( ).size ( );
        size_t lastPrefixPos = posInPattern;

        while ( posInPattern > 0 && currentState != failState ) {
            auto transition = suffixAutomaton.getTransitions ( ).find ( { currentState, subject.getContent ( ).at ( posInSubject + posInPattern - 1 ) } );

            if ( transition == suffixAutomaton.getTransitions ( ).end ( ) )
                currentState = failState;
            else
                currentState = transition->second;

            posInPattern--;

             // found a prefix of nonreversed pattern that does not correspond to the entire pattern
            if ( ( posInPattern != 0 ) && ( suffixAutomaton.getFinalStates ( ).find ( currentState ) != suffixAutomaton.getFinalStates ( ).end ( ) ) )
                lastPrefixPos = posInPattern;
        }

        if ( currentState != failState )
             // Yay, there is match!!!
            occ.insert ( posInSubject );

        posInSubject += lastPrefixPos;
    }

    measurements::end ( );

    return occ;
}

auto BackwardDAWGMatchingLinearStringLinearString = BackwardDAWGMatching::RegistratorWrapper < std::set < unsigned >, string::LinearString < >, string::LinearString < > > ( BackwardDAWGMatching::match );

} /* namespace exact */

} /* namespace stringology */
