/*
 * Author: Radovan Cerveny
 */

#include "SuffixAutomaton.h"
#include <string/Epsilon.h>
#include <label/LabelSetLabel.h>
#include <automaton/determinize/Determinize.h>
#include <automaton/simplify/Minimize.h>
#include <automaton/simplify/EpsilonRemoverIncoming.h>

namespace stringology {

namespace exact {

automaton::Automaton SuffixAutomaton::naiveConstruct ( const string::String & pattern ) {
    return dispatch ( pattern.getData ( ) );
}

automaton::DFA<> SuffixAutomaton::naiveConstruct ( const string::LinearString < > & pattern ) {
    automaton::EpsilonNFA < > nfaSuffixAutomaton ( label::Label ( 0 ) );

    nfaSuffixAutomaton.setInputAlphabet ( pattern.getAlphabet ( ) );

    int i = 0;

    for ( const alphabet::Symbol & symbol : pattern.getContent ( ) ) {
        i++;
        nfaSuffixAutomaton.addState ( label::Label ( i ) );
        nfaSuffixAutomaton.addTransition ( label::Label ( i - 1 ), symbol, label::Label ( i ) );
        nfaSuffixAutomaton.addTransition ( label::Label ( 0 ), label::Label ( i ) );
    }

    nfaSuffixAutomaton.addFinalState ( label::Label ( i ) );

    automaton::DFA<> minimalSuffixAutomaton = automaton::simplify::Minimize::minimize ( automaton::determinize::Determinize::determinize ( automaton::simplify::EpsilonRemoverIncoming::remove ( nfaSuffixAutomaton ) ) );

    // FIXME Trim? and is this even supposed to be here?
    label::Label failState = label::Label ( label::LabelSetLabel ( { } ) );

    auto transitionsToFailState = minimalSuffixAutomaton.getTransitionsToState ( failState );

    for ( const auto & transition : transitionsToFailState )
        minimalSuffixAutomaton.removeTransition ( transition.first.first, transition.first.second, transition.second );

    minimalSuffixAutomaton.removeState ( failState );

    return minimalSuffixAutomaton;
}

automaton::Automaton SuffixAutomaton::construct ( const string::String & pattern ) {
    return dispatch ( pattern.getData ( ) );
}

automaton::DFA<> SuffixAutomaton::construct ( const string::LinearString < > & pattern ) {
    automaton::DFA<> suffixAutomaton ( label::Label ( 0 ) );

    suffixAutomaton.setInputAlphabet ( pattern.getAlphabet ( ) );

    std::map < label::Label, std::pair < label::Label, int > > suffixLinks = { { label::Label ( 0 ), { label::Label ( -1 ), 0 } } };
    label::Label lastState = label::Label ( 0 );

    for ( const alphabet::Symbol & symbol : pattern.getContent ( ) )
        suffixAutomatonAddSymbol ( suffixAutomaton, symbol, suffixLinks, lastState );

    while ( lastState != label::Label ( -1 ) ) {
        suffixAutomaton.addFinalState ( lastState );
        lastState = suffixLinks.find ( lastState )->second.first;
    }

    return suffixAutomaton;
}

void SuffixAutomaton::suffixAutomatonAddSymbol ( automaton::DFA<> & suffixAutomaton, const alphabet::Symbol & symbol, std::map < label::Label, std::pair < label::Label, int > > & suffixLinks, label::Label & lastState ) {

    label::Label newState = label::Label ( ( int ) suffixAutomaton.getStates ( ).size ( ) );

    suffixAutomaton.addState ( newState );

    int lastSuffixLength = suffixLinks.find ( lastState )->second.second;

    suffixLinks.insert ( { newState, { label::Label ( -1 ), lastSuffixLength + 1 } } );

    label::Label kState = lastState;

    while ( kState != label::Label ( -1 ) && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } ) == suffixAutomaton.getTransitions ( ).end ( ) ) {
        suffixAutomaton.addTransition ( kState, symbol, newState );
        kState = suffixLinks.find ( kState )->second.first;
    }

    if ( kState == label::Label ( -1 ) ) {
        suffixLinks.find ( newState )->second.first = label::Label ( 0 );
    } else {
        label::Label qState = suffixAutomaton.getTransitions ( ).find ( { kState, symbol } )->second;

        int kSuffixLength = suffixLinks.find ( kState )->second.second;
        int qSuffixLength = suffixLinks.find ( qState )->second.second;

        if ( kSuffixLength + 1 == qSuffixLength ) {
            suffixLinks.find ( newState )->second.first = qState;
        } else {

            label::Label cloneState = label::Label ( ( int ) suffixAutomaton.getStates ( ).size ( ) );
            suffixAutomaton.addState ( cloneState );

            suffixLinks.insert ( { cloneState, { suffixLinks.find ( qState )->second.first, kSuffixLength + 1 } } );

            for ( const auto & transition : suffixAutomaton.getTransitionsFromState ( qState ) )
                suffixAutomaton.addTransition ( cloneState, transition.first.second, transition.second );

            while ( kState != label::Label ( -1 )
                && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } ) != suffixAutomaton.getTransitions ( ).end ( )
                && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } )->second == qState ) {
                suffixAutomaton.removeTransition ( kState, symbol, qState );
                suffixAutomaton.addTransition ( kState, symbol, cloneState );
                kState = suffixLinks.find ( kState )->second.first;
            }

            suffixLinks.find ( qState )->second.first = cloneState;
            suffixLinks.find ( newState )->second.first = cloneState;
        }
    }
    lastState = newState;
}

auto SuffixAutomatonLinearString = SuffixAutomaton::RegistratorWrapper < automaton::DFA<>, string::LinearString < > > ( SuffixAutomaton::construct );

} /* namespace exact */

} /* namespace stringology */
