/*
 * WrapperBase.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef WRAPPER_BASE_H_
#define WRAPPER_BASE_H_

#include <ostream>
#include <memory>

namespace alib {

class WrapperBaseBase {
public:
	virtual ~WrapperBaseBase ( ) noexcept { }

};

template < typename T >
class WrapperBase : public WrapperBaseBase {
protected:
	std::cow_shared_ptr < T > m_data;

private:
	void unify ( WrapperBase & other ) {
		if ( this->m_data.getUseCount ( ) > other.m_data.getUseCount ( ) )
			other.m_data = this->m_data;
		else
			this->m_data = other.m_data;
	}

public:
	WrapperBase ( int ) = delete;

	explicit WrapperBase ( T * data ) : m_data ( data ) {
	}

	explicit WrapperBase ( const T & data ) : m_data ( data.clone ( ) ) {
	}

	explicit WrapperBase ( T && data ) : m_data ( std::move ( data ).plunder ( ) ) {
	}

	const T & getData ( ) const {
		return * m_data;
	}

	T & getData ( ) {
		return * m_data;
	}

	void setData ( T * data ) {
		this->m_data = std::cow_shared_ptr < T > ( data );
	}

	void setData ( const T & data ) {
		this->m_data = std::cow_shared_ptr < T > ( data.clone ( ) );
	}

	void setData ( T && data ) {
		this->m_data = std::cow_shared_ptr < T > ( std::move ( data ).plunder ( ) );
	}

	bool operator >=( const WrapperBase & other ) const {
		return this->compare ( other ) >= 0;
	}

	bool operator <=( const WrapperBase & other ) const {
		return this->compare ( other ) <= 0;
	}

	bool operator >( const WrapperBase & other ) const {
		return this->compare ( other ) > 0;
	}

	bool operator <( const WrapperBase & other ) const {
		return this->compare ( other ) < 0;
	}

	bool operator !=( const WrapperBase & other ) const {
		return this->compare ( other ) != 0;
	}

	bool operator ==( const WrapperBase & other ) const {
		return this->compare ( other ) == 0;
	}

	int compare ( const WrapperBase & other ) const {
		if ( this->m_data.get ( ) == other.m_data.get ( ) ) return 0;

		int res = ( * this->m_data ).compare ( * other.m_data );

		if ( res == 0 ) const_cast < WrapperBase * > ( this )->unify ( const_cast < WrapperBase & > ( other ) );

		return res;
	}

	friend std::ostream & operator <<( std::ostream & os, const WrapperBase & instance ) {
		os << * ( instance.m_data );
		return os;
	}

	explicit operator std::string ( ) const {
		return ( std::string ) * m_data;
	}

	WrapperBase < T > & operator ++ ( ) {
		T * res = std::move ( this->getData ( ) ).inc ( );

		if ( res != NULL )
			this->setData ( res );

		return *this;
	}

};

} /* namespace alib */

namespace std {

template < class T >
struct compare < T, typename std::enable_if < std::is_base_of < alib::WrapperBaseBase, T >::value >::type > {
	int operator ()( const T & first, const T & second ) const {
		return first.getData ( ).compare ( second.getData ( ) );
	}

};

} /* namespace std */

#endif /* WRAPPER_BASE_H_ */
