/*
 * Map.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_MAP_H_
#define OBJECTS_MAP_H_

#include <map>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"

#include "ObjectsPair.h"

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class KeyType, class ValueType >
class ObjectsMap : public std::map < KeyType, ValueType >, public ContainerBase {
public:
	explicit ObjectsMap ( std::map < KeyType, ValueType > );
	explicit ObjectsMap ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsMap & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Map";

		return xmlTagName;
	}

	static ObjectsMap parse ( std::deque < sax::Token >::iterator & input );
	static std::map < KeyType, ValueType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void compose ( std::deque < sax::Token > & out, const std::map < KeyType, ValueType > & input );

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::ObjectsMap ( std::map < KeyType, ValueType > raw ) : std::map < KeyType, ValueType > ( std::move ( raw ) ) {

}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::ObjectsMap() : std::map<KeyType, ValueType>() {

}

template < class KeyType, class ValueType >
ContainerBase* ObjectsMap<KeyType, ValueType>::clone() const {
	return new ObjectsMap(*this);
}

template < class KeyType, class ValueType >
ContainerBase* ObjectsMap<KeyType, ValueType>::plunder() && {
	return new ObjectsMap(std::move(*this));
}

template < class KeyType, class ValueType >
int ObjectsMap<KeyType, ValueType>::compare(const ObjectsMap& other) const {
	std::compare<std::map<KeyType, ValueType>> comp;
	return comp ( static_cast < const std::map < KeyType, ValueType > & > ( * this ), static_cast < const std::map < KeyType, ValueType > & > ( other ) );
}

template < class KeyType, class ValueType >
void ObjectsMap<KeyType, ValueType>::operator>>(std::ostream& os) const {
	os << "(ObjectsMap " << static_cast < const std::map < KeyType, ValueType > & > ( * this ) << ")";
}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType> ObjectsMap<KeyType, ValueType>::parse(std::deque<sax::Token>::iterator& input) {
	return ObjectsMap<KeyType, ValueType>(parseRaw(input));
}

template < class KeyType, class ValueType >
void ObjectsMap<KeyType, ValueType>::compose(std::deque<sax::Token>& out) const {
	compose(out, *this);
}

template < class KeyType, class ValueType >
std::map < KeyType, ValueType > ObjectsMap < KeyType, ValueType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsMap < >::getXmlTagName() );

	std::map < KeyType, ValueType > map;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		map.insert ( alib::xmlApi < std::pair < KeyType, ValueType > >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsMap < >::getXmlTagName() );
	return map;
}

template < class KeyType, class ValueType >
void ObjectsMap < KeyType, ValueType >::compose ( std::deque < sax::Token > & out, const std::map < KeyType, ValueType > & input ) {
	out.emplace_back ( ObjectsMap < >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	for ( const std::pair < const KeyType, ValueType > & item : input )
		alib::xmlApi < std::pair < KeyType, ValueType > >::compose ( out, item );

	out.emplace_back ( ObjectsMap < >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class KeyType, class ValueType >
alib::ObjectBase* ObjectsMap < KeyType, ValueType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < std::map < T, R > > {
	static std::map < T, R > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const std::map < T, R > & data );
};

template < typename T, typename R >
std::map < T, R > xmlApi < std::map < T, R > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsMap < T, R >::parseRaw ( input );
}

template < typename T, typename R >
bool xmlApi < std::map < T, R > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, container::ObjectsMap < >::getXmlTagName() );
}

template < typename T, typename R >
void xmlApi < std::map < T, R > >::compose ( std::deque < sax::Token > & output, const std::map < T, R > & input ) {
	return container::ObjectsMap < T, R >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_MAP_H_ */
