/*
 * Set.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_SET_H_
#define OBJECTS_SET_H_

#include <set>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class ElementType >
class ObjectsSet : public std::set < ElementType >, public ContainerBase {
public:
	explicit ObjectsSet ( std::set < ElementType > );
	explicit ObjectsSet ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsSet & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Set";

		return xmlTagName;
	}

	static ObjectsSet parse ( std::deque < sax::Token >::iterator & input );
	static std::set < ElementType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void compose ( std::deque < sax::Token > & out, const std::set < ElementType > & input );

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class ElementType >
ObjectsSet < ElementType >::ObjectsSet(std::set<ElementType> raw) : std::set<ElementType>(std::move(raw)) {

}

template < class ElementType >
ObjectsSet < ElementType >::ObjectsSet() : std::set<ElementType>() {

}

template < class ElementType >
ContainerBase* ObjectsSet < ElementType >::clone() const {
	return new ObjectsSet(*this);
}

template < class ElementType >
ContainerBase* ObjectsSet < ElementType >::plunder() && {
	return new ObjectsSet(std::move(*this));
}

template < class ElementType >
int ObjectsSet < ElementType >::compare(const ObjectsSet& other) const {
	std::compare<std::set<ElementType>> comp;
	return comp ( static_cast < const std::set < ElementType > & > ( * this ), static_cast < const std::set < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsSet < ElementType >::operator>>(std::ostream& os) const {
	os << "(ObjectsSet " << static_cast < const std::set < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsSet < ElementType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class ElementType >
ObjectsSet < ElementType > ObjectsSet < ElementType >::parse(std::deque<sax::Token>::iterator& input) {
	return ObjectsSet < ElementType >(parseRaw(input));
}

template < class ElementType >
void ObjectsSet < ElementType >::compose(std::deque<sax::Token>& out) const {
	compose(out, *this);
}

template < class ElementType >
std::set < ElementType > ObjectsSet < ElementType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsSet < >::getXmlTagName() );

	std::set < ElementType > set;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		set.insert ( alib::xmlApi < ElementType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsSet < >::getXmlTagName() );
	return set;
}

template < class ElementType >
void ObjectsSet < ElementType >::compose ( std::deque < sax::Token > & out, const std::set < ElementType > & input ) {
	out.emplace_back ( ObjectsSet < >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	for ( const ElementType & item : input )
		alib::xmlApi < ElementType >::compose ( out, item );

	out.emplace_back ( ObjectsSet < >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class ElementType >
alib::ObjectBase* ObjectsSet < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < std::set < T > > {
	static std::set < T > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const std::set < T > & data );
};

template < typename T >
std::set < T > xmlApi < std::set < T > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsSet < T >::parseRaw ( input );
}

template < typename T >
bool xmlApi < std::set < T > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, container::ObjectsSet < >::getXmlTagName() );
}

template < typename T >
void xmlApi < std::set < T > >::compose ( std::deque < sax::Token > & output, const std::set < T > & input ) {
	return container::ObjectsSet < T >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_SET_H_ */
