/*
 * vector.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_VECTOR_H_
#define OBJECTS_VECTOR_H_

#include <vector>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class ElementType >
class ObjectsVector : public std::vector < ElementType >, public ContainerBase {
public:
	explicit ObjectsVector ( std::vector < ElementType > );
	explicit ObjectsVector ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsVector & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Vector";

		return xmlTagName;
	}

	static ObjectsVector parse ( std::deque < sax::Token >::iterator & input );
	static std::vector < ElementType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void compose ( std::deque < sax::Token > & out, const std::vector < ElementType > & input );

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class ElementType >
ObjectsVector < ElementType >::ObjectsVector ( std::vector < ElementType > raw ) : std::vector < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ObjectsVector < ElementType >::ObjectsVector ( ) : std::vector < ElementType > ( ) {
}

template < class ElementType >
ContainerBase * ObjectsVector < ElementType >::clone ( ) const {
	return new ObjectsVector ( * this );
}

template < class ElementType >
ContainerBase * ObjectsVector < ElementType >::plunder ( ) && {
	return new ObjectsVector ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsVector < ElementType >::compare ( const ObjectsVector & other ) const {
	std::compare < std::vector < ElementType > > comp;

	return comp ( static_cast < const std::vector < ElementType > & > ( * this ), static_cast < const std::vector < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsVector < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsVector " << static_cast < const std::vector < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsVector < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
ObjectsVector < ElementType > ObjectsVector < ElementType >::parse ( std::deque < sax::Token >::iterator & input ) {
	return ObjectsVector < ElementType > ( parseRaw ( input ) );
}

template < class ElementType >
void ObjectsVector < ElementType >::compose ( std::deque < sax::Token > & out ) const {
	compose ( out, * this );
}

template < class ElementType >
std::vector < ElementType > ObjectsVector < ElementType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsVector < >::getXmlTagName() );

	std::vector < ElementType > vector;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		vector.push_back ( alib::xmlApi < ElementType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsVector < >::getXmlTagName() );
	return vector;
}

template < class ElementType >
void ObjectsVector < ElementType >::compose ( std::deque < sax::Token > & out, const std::vector < ElementType > & input ) {
	out.emplace_back ( ObjectsVector < >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	for ( const ElementType & item : input )
		alib::xmlApi < ElementType >::compose ( out, item );

	out.emplace_back ( ObjectsVector < >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class ElementType >
alib::ObjectBase* ObjectsVector < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < std::vector < T > > {
	static std::vector < T > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const std::vector < T > & data );
};

template < typename T >
std::vector < T > xmlApi < std::vector < T > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsVector < T >::parseRaw ( input );
}

template < typename T >
bool xmlApi < std::vector < T > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, container::ObjectsVector < >::getXmlTagName() );
}

template < typename T >
void xmlApi < std::vector < T > >::compose ( std::deque < sax::Token > & output, const std::vector < T > & input ) {
	return container::ObjectsVector < T >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_VECTOR_H_ */
