/*
 * DPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef DPDA_H_
#define DPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Automaton
 *
 * $|\delta (q, a, \gamma)| \leq 1$, $\forall q, a, \gamma, q \in Q, a \in (T \cup \{\varepsilon\}), \gamma \in G^*.$
 * if $\delta (q, a, \alpha) \neq \emptyset$, $\delta (q, a, \beta) \neq \emptyset$ and $\alpha \neq \beta$, then $\alpha$ is not suffix of $\beta$ and $\beta$ is not suffix of $\alpha$ (formally $\gamma \alpha \neq \beta and \alpha \neq \gamma \beta$).
 * if $\delta(q, a, \alpha) \neq \emptyset$, $\delta (q, \varepsilon, \beta) \neq \emptyset$, then $\alpha$ is not suffix of $\beta$ and $\beta$ is not suffix of $\alpha$ (fornally $\gamma \alpha \neq \beta and \alpha \neq \gamma \beta$).
 */
template < class SymbolType, class EpsilonType, class StateType >
class DPDA : public AutomatonBase, public std::Components < DPDA < SymbolType, EpsilonType, StateType >, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::pair < StateType, std::vector < SymbolType > > > transitions;

public:
	explicit DPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreSymbol, StateType initialState, SymbolType initialPushdownSymbol, std::set < StateType > finalStates );
	explicit DPDA ( StateType initialState, SymbolType initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the DPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the DPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, std::variant < EpsilonType, SymbolType > input, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push );

	bool addTransition ( StateType from, SymbolType input, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push );

	bool addTransition ( StateType from, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push );

	/**
	 * Removes the transition from the DPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the DPDA
	 */
	bool removeTransition ( const StateType & from, const std::variant < EpsilonType, SymbolType > & input, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push );

	bool removeTransition ( const StateType & from, const SymbolType & input, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push );

	bool removeTransition ( const StateType & from, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push );

	/**
	 * @return DPDA transitions
	 */
	const std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::pair < StateType, std::vector < SymbolType > > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::pair < StateType, std::vector < SymbolType > > > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::pair < StateType, std::vector < SymbolType > > > getTransitionsToState ( const StateType & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const DPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "DPDA";

		return xmlTagName;
	}

	static DPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, DPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class EpsilonType, class StateType >
DPDA < SymbolType, EpsilonType, StateType >::DPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreAlphabet, StateType initialState, SymbolType initialSymbol, std::set < StateType > finalStates ) : std::Components < DPDA, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class SymbolType, class EpsilonType, class StateType >
DPDA < SymbolType, EpsilonType, StateType >::DPDA(StateType initialState, SymbolType initialPushdownSymbol) : DPDA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, std::set < SymbolType > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* DPDA < SymbolType, EpsilonType, StateType >::clone() const {
	return new DPDA(*this);
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* DPDA < SymbolType, EpsilonType, StateType >::plunder() && {
	return new DPDA(std::move(*this));
}

template < class SymbolType, class EpsilonType, class StateType >
bool DPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::variant<EpsilonType, SymbolType> input, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is<SymbolType>() && ! getInputAlphabet().count(input.template get<SymbolType>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	for(const SymbolType& popSymbol : pop) {
		if (! getPushdownStoreAlphabet().count(popSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( popSymbol ) + "\" doesn't exist.");
		}
	}

	for(const SymbolType& pushSymbol : push) {
		if (! getPushdownStoreAlphabet().count(pushSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");
		}
	}

	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> > key(std::move(from), std::move(input), std::move(pop));
	std::pair<StateType, std::vector<SymbolType> > value = std::make_pair(std::move(to), std::move(push));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (std::get<1>(key).template is<SymbolType>())
			throw AutomatonException(
				"Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).template get<SymbolType>() ) + "\", \"pop\") -> ?? already exists.");
		else
			throw AutomatonException(
				"Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).template get<EpsilonType>() ) + "\", \"pop\") -> ?? already exists.");
	}

	if(std::get<1>(key).template is<EpsilonType>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key)) {
					const std::vector<SymbolType>& alpha = std::get<2>(transition.first);
					const std::vector<SymbolType>& beta = std::get<2>(key);

					const std::vector<SymbolType>& shorter = (alpha.size() < beta.size()) ? alpha : beta;
					const std::vector<SymbolType>& longer = (alpha.size() < beta.size()) ? beta : alpha;

					for(auto itS = shorter.begin(), itL = longer.begin(); itS != shorter.end(); ++itS, ++itL) {
						if(*itS != *itL) return false;
					}
					return true;
				}
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	} else {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key) && ( std::get<1>(transition.first) == std::get<1>(key) || std::get<1>(transition.first).template is<EpsilonType>() )) {
					const std::vector<SymbolType>& alpha = std::get<2>(transition.first);
					const std::vector<SymbolType>& beta = std::get<2>(key);

					const std::vector<SymbolType>& shorter = (alpha.size() < beta.size()) ? alpha : beta;
					const std::vector<SymbolType>& longer = (alpha.size() < beta.size()) ? beta : alpha;

					for(auto itS = shorter.begin(), itL = longer.begin(); itS != shorter.end(); ++itS, ++itL) {
						if(*itS != *itL) return false;
					}
					return true;
				}
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)) );
	return true;
}

template < class SymbolType, class EpsilonType, class StateType >
bool DPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, SymbolType input, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push) {
	std::variant<EpsilonType, SymbolType> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool DPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool DPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push) {
	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> > key(from, input, pop);
	std::pair<StateType, std::vector<SymbolType> > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.template is<SymbolType>())
			throw AutomatonException(
				"Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.template get<SymbolType>() ) + "\") -> \"to\" doesn't exist.");
		else
			throw AutomatonException(
				"Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.template get<EpsilonType>() ) + "\") -> \"to\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

template < class SymbolType, class EpsilonType, class StateType >
bool DPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const SymbolType& input, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
bool DPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > >& DPDA < SymbolType, EpsilonType, StateType >::getTransitions() const {
	return transitions;
}

template < class SymbolType, class EpsilonType, class StateType >
std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > > DPDA < SymbolType, EpsilonType, StateType >::getTransitionsFromState(const StateType& from) const {
	if( ! getStates().count(from) )
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist");

	std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > > transitionsFromState;
	for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > >& transition : transitions) {
		if (std::get<0>(transition.first) == from) {
			transitionsFromState.insert(std::make_pair(transition.first, transition.second));
		}
	}

	return transitionsFromState;
}

template < class SymbolType, class EpsilonType, class StateType >
std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > > DPDA < SymbolType, EpsilonType, StateType >::getTransitionsToState(const StateType& to) const {
	if( ! getStates().count(to))
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist");

	std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > > transitionsToState;
	for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > >& transition : transitions) {
		if (transition.second.first == to) {
			transitionsToState.insert(std::make_pair(transition.first, transition.second));
		}
	}

	return transitionsToState;
}

template < class SymbolType, class EpsilonType, class StateType >
int DPDA < SymbolType, EpsilonType, StateType >::compare(const DPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class EpsilonType, class StateType >
void DPDA < SymbolType, EpsilonType, StateType >::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

template < class SymbolType, class EpsilonType, class StateType >
DPDA < SymbolType, EpsilonType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class EpsilonType, class StateType >
DPDA < SymbolType, EpsilonType, StateType > DPDA < SymbolType, EpsilonType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<SymbolType>(input);
	std::set<SymbolType> pushdownStoreSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	SymbolType initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	DPDA automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(pushdownStoreSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DPDA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class EpsilonType, class StateType >
void DPDA < SymbolType, EpsilonType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, DPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
	std::vector<SymbolType> pop = AutomatonFromXMLParser::parseTransitionPop<SymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	std::vector<SymbolType> push = AutomatonFromXMLParser::parseTransitionPush<SymbolType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
void DPDA < SymbolType, EpsilonType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
void DPDA < SymbolType, EpsilonType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionPop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionPush(out, transition.second.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
alib::ObjectBase* DPDA < SymbolType, EpsilonType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::DPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).template is<SymbolType>() && symbol == std::get<1>(transition.first).template get<SymbolType>())
				return true;

		return false;
	}

	static bool available ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::DPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getInitialSymbol() == symbol)
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > >& transition : automaton.getTransitions()) {
			const auto & popSymbols = std::get<2>(transition.first);
			const auto & pushSymbols = transition.second.second;
			if(std::contains(popSymbols.begin(), popSymbols.end(), symbol ) || std::contains(pushSymbols.begin(), pushSymbols.end(), symbol))
				return true;
		}

		return false;
	}

	static bool available ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::DPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::DPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::pair<StateType, std::vector<SymbolType> > >& transition : automaton.getTransitions())
			if ( state == std::get<0>(transition.first) || transition.second.first == state )
				return true;

		return false;
	}

	static bool available ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::DPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::DPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::DPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* DPDA_H_ */
