/*
 * NPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef NPDA_H_
#define NPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Automaton
 */
template < class SymbolType, class EpsilonType, class StateType >
class NPDA : public AutomatonBase, public std::Components < NPDA < SymbolType, EpsilonType, StateType >, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::set < std::pair < StateType, std::vector < SymbolType > > > > transitions;

public:
	explicit NPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreSymbol, StateType initialState, SymbolType initialPushdownSymbol, std::set < StateType > finalStates );
	explicit NPDA ( StateType initialState, SymbolType initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the NPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the NPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, std::variant < EpsilonType, SymbolType > input, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push );

	bool addTransition ( StateType from, SymbolType input, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push );

	bool addTransition ( StateType from, std::vector < SymbolType > pop, StateType to, std::vector < SymbolType > push );

	/**
	 * Removes the transition from the NPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the NPDA
	 */
	bool removeTransition ( const StateType & from, const std::variant < EpsilonType, SymbolType > & input, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push );

	bool removeTransition ( const StateType & from, const SymbolType & input, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push );

	bool removeTransition ( const StateType & from, const std::vector < SymbolType > & pop, const StateType & to, const std::vector < SymbolType > & push );

	/**
	 * @return NPDA transitions
	 */
	const std::map < std::tuple < StateType, std::variant < EpsilonType, SymbolType >, std::vector < SymbolType > >, std::set < std::pair < StateType, std::vector < SymbolType > > > > & getTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "NPDA";

		return xmlTagName;
	}

	static NPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, NPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class EpsilonType, class StateType >
NPDA < SymbolType, EpsilonType, StateType >::NPDA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < SymbolType > pushdownStoreAlphabet, StateType initialState, SymbolType initialSymbol, std::set < StateType > finalStates ) : std::Components < NPDA, SymbolType, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class SymbolType, class EpsilonType, class StateType >
NPDA < SymbolType, EpsilonType, StateType >::NPDA(StateType initialState, SymbolType initialPushdownSymbol) : NPDA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, std::set < SymbolType > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* NPDA < SymbolType, EpsilonType, StateType >::clone() const {
	return new NPDA(*this);
}

template < class SymbolType, class EpsilonType, class StateType >
AutomatonBase* NPDA < SymbolType, EpsilonType, StateType >::plunder() && {
	return new NPDA(std::move(*this));
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::variant<EpsilonType, SymbolType> input, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push) {
	if (! getStates().count(from))
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");

	if (input.template is<SymbolType>() && ! getInputAlphabet().count(input.template get<SymbolType>()))
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get<SymbolType>() ) + "\" doesn't exist.");

	if (! getStates().count(to))
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");

	for(const SymbolType& popSymbol : pop)
		if (! getPushdownStoreAlphabet().count(popSymbol))
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( popSymbol ) + "\" doesn't exist.");

	for(const SymbolType& pushSymbol : push)
		if (! getPushdownStoreAlphabet().count(pushSymbol))
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");

	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> > key(std::move(from), std::move(input), std::move(pop));
	std::pair<StateType, std::vector<SymbolType> > value = std::make_pair(std::move(to), std::move(push));

	return transitions[std::move(key)].insert(std::move(value)).second;
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, SymbolType input, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push) {
	std::variant<EpsilonType, SymbolType> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDA < SymbolType, EpsilonType, StateType >::addTransition(StateType from, std::vector<SymbolType> pop, StateType to, std::vector<SymbolType> push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::variant<EpsilonType, SymbolType>& input, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push) {
	std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> > key(from, input, pop);
	std::pair<StateType, std::vector<SymbolType> > value = std::make_pair(to, push);

	return transitions[key].erase(value);
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const SymbolType& input, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push) {
	std::variant<EpsilonType, SymbolType> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
bool NPDA < SymbolType, EpsilonType, StateType >::removeTransition(const StateType& from, const std::vector<SymbolType>& pop, const StateType& to, const std::vector<SymbolType>& push) {
	auto inputVariant = std::variant<EpsilonType, SymbolType>::template from < EpsilonType > ( );
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class SymbolType, class EpsilonType, class StateType >
const std::map<std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::pair<StateType, std::vector<SymbolType> > > >& NPDA < SymbolType, EpsilonType, StateType >::getTransitions() const {
	return transitions;
}

template < class SymbolType, class EpsilonType, class StateType >
int NPDA < SymbolType, EpsilonType, StateType >::compare(const NPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDA < SymbolType, EpsilonType, StateType >::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

template < class SymbolType, class EpsilonType, class StateType >
NPDA < SymbolType, EpsilonType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class EpsilonType, class StateType >
NPDA < SymbolType, EpsilonType, StateType > NPDA < SymbolType, EpsilonType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<SymbolType>(input);
	std::set<SymbolType> pushdownStoreSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	SymbolType initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	NPDA automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(pushdownStoreSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<NPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NPDA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDA < SymbolType, EpsilonType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, NPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	std::variant<EpsilonType, SymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, SymbolType>(input);
	std::vector<SymbolType> pop = AutomatonFromXMLParser::parseTransitionPop<SymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	std::vector<SymbolType> push = AutomatonFromXMLParser::parseTransitionPush<SymbolType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDA < SymbolType, EpsilonType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(NPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
void NPDA < SymbolType, EpsilonType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionPop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionPush(out, target.second);

			out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class EpsilonType, class StateType >
alib::ObjectBase* NPDA < SymbolType, EpsilonType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::NPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::pair<StateType, std::vector<SymbolType> > > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).template is<SymbolType>() && symbol == std::get<1>(transition.first).template get<SymbolType>())
				return true;

		return false;
	}

	static bool available ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::NPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getInitialSymbol() == symbol)
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::pair<StateType, std::vector<SymbolType> > > >& transition : automaton.getTransitions()) {
			for (const SymbolType& popSymbol : std::get<2>(transition.first))
				if (symbol == popSymbol)
					return true;

			for (const std::pair<StateType, std::vector<SymbolType> >& target : transition.second)
				if (std::find(target.second.begin(), target.second.end(), symbol) != target.second.end())
					return true;
		}

		return false;
	}

	static bool available ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::NPDA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::NPDA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::NPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, SymbolType>, std::vector<SymbolType> >, std::set<std::pair<StateType, std::vector<SymbolType> > > >& transition : automaton.getTransitions()) {
			if (state == std::get<0>(transition.first))
				return true;
			for(const std::pair<StateType, std::vector<SymbolType> >& target : transition.second)
				if(target.first == state)
					return true;
		}

		return false;
	}

	static bool available ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::NPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::NPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::NPDA < SymbolType, EpsilonType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::NPDA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NPDA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* NPDA_H_ */
