/*
 * VisiblyPushdownDPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef VISIBLY_PUSHDOWN_DPDA_H_
#define VISIBLY_PUSHDOWN_DPDA_H_

#include <map>
#include <vector>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class CallAlphabet;
class ReturnAlphabet;
class LocalAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
template < class SymbolType, class StateType >
class VisiblyPushdownDPDA : public AutomatonBase, public std::Components < VisiblyPushdownDPDA < SymbolType, StateType >, SymbolType, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < StateType, SymbolType >, std::pair < StateType, SymbolType > > callTransitions;
	std::map < std::tuple < StateType, SymbolType, SymbolType >, StateType > returnTransitions;
	std::map < std::pair < StateType, SymbolType >, StateType > localTransitions;

public:
	explicit VisiblyPushdownDPDA ( std::set < StateType > states, std::set < SymbolType > callAlphabet, std::set < SymbolType > returnAlphabet, std::set < SymbolType > localAlphabet, std::set < SymbolType > pushdownStoreSymbol, StateType initialState, SymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates );
	explicit VisiblyPushdownDPDA ( StateType initialSymbol, SymbolType bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	bool removeState ( const StateType & state ) {
		return this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	bool removeFinalState ( const StateType & state ) {
		return this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removePushdownStoreSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getBottomOfTheStackSymbol ( ) const {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( SymbolType symbol ) {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getCallInputAlphabet ( ) const {
		return this->template accessComponent < CallAlphabet > ( ).get ( );
	}

	bool addCallInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < CallAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addCallInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < CallAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setCallInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < CallAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeCallInputSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < CallAlphabet > ( ).remove ( symbol );
	}

	const std::set < SymbolType > & getReturnInputAlphabet ( ) const {
		return this->template accessComponent < ReturnAlphabet > ( ).get ( );
	}

	bool addReturnInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addReturnInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setReturnInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < ReturnAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeReturnInputSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < ReturnAlphabet > ( ).remove ( symbol );
	}

	const std::set < SymbolType > & getLocalInputAlphabet ( ) const {
		return this->template accessComponent < LocalAlphabet > ( ).get ( );
	}

	bool addLocalInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < LocalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addLocalInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < LocalAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setLocalInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < LocalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeLocalInputSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < LocalAlphabet > ( ).remove ( symbol );
	}

	bool removeInputSymbol(const SymbolType& symbol) {
		return this->template removeCallInputSymbol(symbol) || removeReturnInputSymbol(symbol) || removeLocalInputSymbol(symbol);
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( StateType current, SymbolType input, StateType next, SymbolType push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( StateType current, SymbolType input, SymbolType pop, StateType next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( StateType current, SymbolType input, StateType next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const StateType & current, const SymbolType & input, const StateType & next, const SymbolType & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const StateType & current, const SymbolType & input, const SymbolType & pop, const StateType & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const StateType & current, const SymbolType & input, const StateType & next );

	const std::map < std::pair < StateType, SymbolType >, std::pair < StateType, SymbolType > > & getCallTransitions ( ) const;

	const std::map < std::tuple < StateType, SymbolType, SymbolType >, StateType > & getReturnTransitions ( ) const;

	const std::map < std::pair < StateType, SymbolType >, StateType > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const VisiblyPushdownDPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "VisiblyPushdownDPDA";

		return xmlTagName;
	}

	static VisiblyPushdownDPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, VisiblyPushdownDPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class StateType >
VisiblyPushdownDPDA < SymbolType, StateType >::VisiblyPushdownDPDA ( std::set < StateType > states, std::set < SymbolType > callAlphabet, std::set < SymbolType > returnAlphabet, std::set < SymbolType > localAlphabet, std::set < SymbolType > pushdownStoreAlphabet, StateType initialState, SymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates ) : std::Components < VisiblyPushdownDPDA, SymbolType, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( callAlphabet ), std::move ( returnAlphabet ), std::move ( localAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( bottomOfTheStackSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class SymbolType, class StateType >
VisiblyPushdownDPDA < SymbolType, StateType >::VisiblyPushdownDPDA(StateType initialState, SymbolType bottomOfTheStackSymbol) : VisiblyPushdownDPDA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, std::set < SymbolType > { }, std::set < SymbolType > { }, std::set < SymbolType > { bottomOfTheStackSymbol }, initialState, bottomOfTheStackSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class StateType >
AutomatonBase* VisiblyPushdownDPDA < SymbolType, StateType >::clone() const {
	return new VisiblyPushdownDPDA(*this);
}

template < class SymbolType, class StateType >
AutomatonBase* VisiblyPushdownDPDA < SymbolType, StateType >::plunder() && {
	return new VisiblyPushdownDPDA(std::move(*this));
}

template < class SymbolType, class StateType >
bool VisiblyPushdownDPDA < SymbolType, StateType >::addCallTransition(StateType from, SymbolType input, StateType to, SymbolType push) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (!getCallInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(push)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( push ) + "\" doesn't exist.");
	}

	std::pair<StateType, SymbolType> key(std::move(from), std::move(input));
	std::pair<StateType, SymbolType> value = std::make_pair(std::move(to), std::move(push));

	if(callTransitions.find(key) != callTransitions.end() && callTransitions.find(key)->second == value)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( key.first ) + "\" when transition reading \"" + std::to_string ( key.second ) + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == key.first && std::get<1>(transition.first) == key.second)
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( key.first ) + "\" when transition reading \"" + std::to_string ( key.second ) + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( key.first ) + "\" when transition reading \"" + std::to_string ( key.second ) + "\" is present.");

	callTransitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownDPDA < SymbolType, StateType >::addReturnTransition(StateType from, SymbolType input, SymbolType pop, StateType to) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (!getReturnInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	std::tuple<StateType, SymbolType, SymbolType> key(std::move(from), std::move(input), std::move(pop));

	if(returnTransitions.find(key) != returnTransitions.end() && returnTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == std::get<0>(key) && transition.first.second == std::get<1>(key))
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( std::get<0>(key) ) + "\" when transition reading \"" + std::to_string ( std::get<1>(key) ) + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == std::get<0>(key) && std::get<1>(transition.first) == std::get<1>(key) && std::get<2>(transition.first) == std::get<2>(key))
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( from ) + "\" when transition reading \"" + std::to_string ( input ) + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == std::get<0>(key) && transition.first.second == std::get<1>(key))
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( std::get<0>(key) ) + "\" when transition reading \"" + std::to_string ( std::get<1>(key) ) + "\" is present.");

	returnTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownDPDA < SymbolType, StateType >::addLocalTransition(StateType from, SymbolType input, StateType to) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (!getLocalInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	std::pair<StateType, SymbolType> key(std::move(from), std::move(input));

	if(localTransitions.find(key) != localTransitions.end() && localTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( key.first ) + "\" when transition reading \"" + std::to_string ( key.second ) + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == key.first && std::get<1>(transition.first) == key.second)
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( key.first ) + "\" when transition reading \"" + std::to_string ( key.second ) + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + std::to_string ( key.first ) + "\" when transition reading \"" + std::to_string ( key.second ) + "\" is present.");

	localTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownDPDA < SymbolType, StateType >::removeCallTransition(const StateType& from, const SymbolType& input, const StateType& to, const SymbolType& push) {
	std::pair<StateType, SymbolType> key(from, input);
	std::pair<StateType, SymbolType> value = std::make_pair(to, push);

	if (callTransitions.find(key) == callTransitions.end())
		return false;

	if(callTransitions.find(key)->second != value)
		throw AutomatonException("Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");

	callTransitions.erase(key);
	return true;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownDPDA < SymbolType, StateType >::removeReturnTransition(const StateType& from, const SymbolType& input, const SymbolType& pop, const StateType& to) {
	std::tuple<StateType, SymbolType, SymbolType> key(from, input, pop);

	if (returnTransitions.find(key) == returnTransitions.end())
		return false;

	if(returnTransitions.find(key)->second != to)
		throw AutomatonException( "Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");

	returnTransitions.erase(key);
	return true;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownDPDA < SymbolType, StateType >::removeLocalTransition(const StateType& from, const SymbolType& input, const StateType& to) {
	std::pair<StateType, SymbolType> key(from, input);

	if (localTransitions.find(key) == localTransitions.end())
		return false;

	if(localTransitions.find(key)->second != to)
		throw AutomatonException("Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");

	localTransitions.erase(key);
	return true;
}

template < class SymbolType, class StateType >
const std::map<std::pair<StateType, SymbolType>, std::pair<StateType, SymbolType> >& VisiblyPushdownDPDA < SymbolType, StateType >::getCallTransitions() const {
	return callTransitions;
}

template < class SymbolType, class StateType >
const std::map<std::tuple<StateType, SymbolType, SymbolType>, StateType>& VisiblyPushdownDPDA < SymbolType, StateType >::getReturnTransitions() const {
	return returnTransitions;
}

template < class SymbolType, class StateType >
const std::map<std::pair<StateType, SymbolType>, StateType>& VisiblyPushdownDPDA < SymbolType, StateType >::getLocalTransitions() const {
	return localTransitions;
}

template < class SymbolType, class StateType >
int VisiblyPushdownDPDA < SymbolType, StateType >::compare(const VisiblyPushdownDPDA& other) const {
	auto first = std::tie(getStates(), getCallInputAlphabet(), getReturnInputAlphabet(), getLocalInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getBottomOfTheStackSymbol(), callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.getStates(), other.getCallInputAlphabet(), other.getReturnInputAlphabet(), other.getLocalInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getBottomOfTheStackSymbol(), other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class StateType >
void VisiblyPushdownDPDA < SymbolType, StateType >::operator>>(std::ostream& out) const {
	out << "(VisiblyPushdownDPDA "
		<< "states = " << getStates()
		<< "callAlphabet = " << getCallInputAlphabet()
		<< "returnAlphabet = " << getReturnInputAlphabet()
		<< "localAlphabet = " << getLocalInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "bottomOfTheStackSymbol = " << getBottomOfTheStackSymbol()
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

template < class SymbolType, class StateType >
VisiblyPushdownDPDA < SymbolType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class StateType >
VisiblyPushdownDPDA < SymbolType, StateType > VisiblyPushdownDPDA < SymbolType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, VisiblyPushdownDPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> callAlphabet = AutomatonFromXMLParser::parseCallInputAlphabet<SymbolType>(input);
	std::set<SymbolType> returnAlphabet = AutomatonFromXMLParser::parseReturnInputAlphabet<SymbolType>(input);
	std::set<SymbolType> localAlphabet = AutomatonFromXMLParser::parseLocalInputAlphabet<SymbolType>(input);
	std::set<SymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	SymbolType bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	VisiblyPushdownDPDA < SymbolType, StateType > automaton(std::move(initialState), std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setCallInputAlphabet(std::move(callAlphabet));
	automaton.setReturnInputAlphabet(std::move(returnAlphabet));
	automaton.setLocalInputAlphabet(std::move(localAlphabet));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<VisiblyPushdownDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, VisiblyPushdownDPDA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class StateType >
void VisiblyPushdownDPDA < SymbolType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, VisiblyPushdownDPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
		SymbolType push = AutomatonFromXMLParser::parseTransitionSinglePush<SymbolType>(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
		SymbolType pop = AutomatonFromXMLParser::parseTransitionSinglePop<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

template < class SymbolType, class StateType >
void VisiblyPushdownDPDA < SymbolType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(VisiblyPushdownDPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeCallInputAlphabet(out, this->getCallInputAlphabet());
	AutomatonToXMLComposer::composeReturnInputAlphabet(out, this->getReturnInputAlphabet());
	AutomatonToXMLComposer::composeLocalInputAlphabet(out, this->getLocalInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(VisiblyPushdownDPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class StateType >
void VisiblyPushdownDPDA < SymbolType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionSinglePush(out, transition.second.second);

		out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getReturnTransitions()) {
		out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getLocalTransitions()) {
		out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class StateType >
alib::ObjectBase* VisiblyPushdownDPDA < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, SymbolType, automaton::CallAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, SymbolType>, std::pair<StateType, SymbolType> >& callTransition : automaton.getCallTransitions())
			if (symbol == callTransition.first.second)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getLocalInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in local alphabet");
		if(automaton.getReturnInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in return alphabet");
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, SymbolType, automaton::ReturnAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, SymbolType, SymbolType>, StateType>& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<1>(returnTransition.first))
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getLocalInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in local alphabet");
		if(automaton.getCallInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in call alphabet");
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, SymbolType, automaton::LocalAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, SymbolType>, StateType>& localTransition : automaton.getLocalTransitions())
			if (symbol == localTransition.first.second)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getReturnInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in return alphabet");
		if(automaton.getCallInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in call alphabet");
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, SymbolType>, std::pair<StateType, SymbolType> >& callTransition : automaton.getCallTransitions())
			if (symbol == callTransition.second.second)
				return true;

		for (const std::pair<const std::tuple<StateType, SymbolType, SymbolType>, StateType>& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<2>(returnTransition.first))
				return true;

		if(automaton.getBottomOfTheStackSymbol() == symbol)
			return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class StateType >
class ElementConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, SymbolType, automaton::BottomOfTheStackSymbol > {
public:
	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<StateType, SymbolType>, std::pair<StateType, SymbolType> >& callTransition : automaton.getCallTransitions())
			if (state == callTransition.first.first || callTransition.second.first == state)
				return true;

		for (const std::pair<const std::tuple<StateType, SymbolType, SymbolType>, StateType>& returnTransition : automaton.getReturnTransitions())
			if (state == std::get<0>(returnTransition.first) || returnTransition.second == state)
				return true;

		for (const std::pair<const std::pair<StateType, SymbolType>, StateType>& localTransition : automaton.getLocalTransitions())
			if (state == localTransition.first.first || localTransition.second == state)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ElementConstraint< automaton::VisiblyPushdownDPDA < SymbolType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::VisiblyPushdownDPDA < SymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* VISIBLY_PUSHDOWN_DPDA_H_ */
