/*
 * VisiblyPushdownNPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef VISIBLY_PUSHDOWN_NPDA_H_
#define VISIBLY_PUSHDOWN_NPDA_H_

#include <map>
#include <vector>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"

namespace automaton {

class CallAlphabet;
class ReturnAlphabet;
class LocalAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialStates;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
template < class SymbolType, class StateType >
class VisiblyPushdownNPDA : public AutomatonBase, public std::Components < VisiblyPushdownNPDA < SymbolType, StateType >, SymbolType, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < StateType, SymbolType >, std::set < std::pair < StateType, SymbolType > > > callTransitions;
	std::map < std::tuple < StateType, SymbolType, SymbolType >, std::set < StateType > > returnTransitions;
	std::map < std::pair < StateType, SymbolType >, std::set < StateType > > localTransitions;

public:
	explicit VisiblyPushdownNPDA ( std::set < StateType > states, std::set < SymbolType > callAlphabet, std::set < SymbolType > returnAlphabet, std::set < SymbolType > localAlphabet, std::set < SymbolType > pushdownStoreSymbol, std::set < StateType > initialStates, SymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates );
	explicit VisiblyPushdownNPDA ( SymbolType bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < StateType > & getInitialStates ( ) const {
		return this->template accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( StateType state ) {
		return this->template accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < StateType > states ) {
		this->template accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	bool removeInitialState ( const StateType & state ) {
		return this->template accessComponent < InitialStates > ( ).remove ( state );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	bool removeState ( const StateType & state ) {
		return this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	bool removeFinalState ( const StateType & state ) {
		return this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( SymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removePushdownStoreSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const SymbolType & getBottomOfTheStackSymbol ( ) const {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( SymbolType symbol ) {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getCallInputAlphabet ( ) const {
		return this->template accessComponent < CallAlphabet > ( ).get ( );
	}

	bool addCallInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < CallAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addCallInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < CallAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setCallInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < CallAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeCallInputSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < CallAlphabet > ( ).remove ( symbol );
	}

	const std::set < SymbolType > & getReturnInputAlphabet ( ) const {
		return this->template accessComponent < ReturnAlphabet > ( ).get ( );
	}

	bool addReturnInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addReturnInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setReturnInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < ReturnAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeReturnInputSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < ReturnAlphabet > ( ).remove ( symbol );
	}

	const std::set < SymbolType > & getLocalInputAlphabet ( ) const {
		return this->template accessComponent < LocalAlphabet > ( ).get ( );
	}

	bool addLocalInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < LocalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addLocalInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < LocalAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setLocalInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < LocalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeLocalInputSymbol ( const SymbolType & symbol ) {
		return this->template accessComponent < LocalAlphabet > ( ).remove ( symbol );
	}

	bool removeInputSymbol(const SymbolType& symbol) {
		return this->template removeCallInputSymbol(symbol) || removeReturnInputSymbol(symbol) || removeLocalInputSymbol(symbol);
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( StateType current, SymbolType input, StateType next, SymbolType push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( StateType current, SymbolType input, SymbolType pop, StateType next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( StateType current, SymbolType input, StateType next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const StateType & current, const SymbolType & input, const StateType & next, const SymbolType & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const StateType & current, const SymbolType & input, const SymbolType & pop, const StateType & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const StateType & current, const SymbolType & input, const StateType & next );

	const std::map < std::pair < StateType, SymbolType >, std::set < std::pair < StateType, SymbolType > > > & getCallTransitions ( ) const;

	const std::map < std::tuple < StateType, SymbolType, SymbolType >, std::set < StateType > > & getReturnTransitions ( ) const;

	const std::map < std::pair < StateType, SymbolType >, std::set < StateType > > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const VisiblyPushdownNPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "VisiblyPushdownNPDA";

		return xmlTagName;
	}

	static VisiblyPushdownNPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, VisiblyPushdownNPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

template < class SymbolType, class StateType >
VisiblyPushdownNPDA < SymbolType, StateType >::VisiblyPushdownNPDA ( std::set < StateType > states, std::set < SymbolType > callAlphabet, std::set < SymbolType > returnAlphabet, std::set < SymbolType > localAlphabet, std::set < SymbolType > pushdownStoreAlphabet, std::set < StateType > initialStates, SymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates ) : std::Components < VisiblyPushdownNPDA, SymbolType, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( callAlphabet ), std::move ( returnAlphabet ), std::move ( localAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( bottomOfTheStackSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( initialStates ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

template < class SymbolType, class StateType >
VisiblyPushdownNPDA < SymbolType, StateType >::VisiblyPushdownNPDA(SymbolType bottomOfTheStackSymbol) : VisiblyPushdownNPDA ( std::set < StateType > { }, std::set < SymbolType > { }, std::set < SymbolType > { }, std::set < SymbolType > { }, std::set < SymbolType > { bottomOfTheStackSymbol }, std::set < StateType > { }, bottomOfTheStackSymbol, std::set < StateType > { }) {
}

template < class SymbolType, class StateType >
AutomatonBase* VisiblyPushdownNPDA < SymbolType, StateType >::clone() const {
	return new VisiblyPushdownNPDA(*this);
}

template < class SymbolType, class StateType >
AutomatonBase* VisiblyPushdownNPDA < SymbolType, StateType >::plunder() && {
	return new VisiblyPushdownNPDA(std::move(*this));
}

template < class SymbolType, class StateType >
bool VisiblyPushdownNPDA < SymbolType, StateType >::addCallTransition(StateType from, SymbolType input, StateType to, SymbolType push) {
	if ( ! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if ( ! getCallInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input ) + "\" doesn't exist.");
	}

	if ( ! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(push)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( push ) + "\" doesn't exist.");
	}

	std::pair<StateType, SymbolType> key(std::move(from), std::move(input));
	std::pair<StateType, SymbolType> value = std::make_pair(std::move(to), std::move(push));

	return callTransitions[std::move(key)].insert(std::move(value)).second;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownNPDA < SymbolType, StateType >::addReturnTransition(StateType from, SymbolType input, SymbolType pop, StateType to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (! getReturnInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	std::tuple<StateType, SymbolType, SymbolType> key(std::move(from), std::move(input), std::move(pop));

	return returnTransitions[std::move(key)].insert(std::move(to)).second;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownNPDA < SymbolType, StateType >::addLocalTransition(StateType from, SymbolType input, StateType to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (! getLocalInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	std::pair<StateType, SymbolType> key(std::move(from), std::move(input));

	return localTransitions[std::move(key)].insert(std::move(to)).second;
}

template < class SymbolType, class StateType >
bool VisiblyPushdownNPDA < SymbolType, StateType >::removeCallTransition(const StateType& from, const SymbolType& input, const StateType& to, const SymbolType& push) {
	std::pair<StateType, SymbolType> key(from, input);
	std::pair<StateType, SymbolType> value = std::make_pair(to, push);

	return callTransitions[key].erase(value);
}

template < class SymbolType, class StateType >
bool VisiblyPushdownNPDA < SymbolType, StateType >::removeReturnTransition(const StateType& from, const SymbolType& input, const SymbolType& pop, const StateType& to) {
	std::tuple<StateType, SymbolType, SymbolType> key(from, input, pop);

	return returnTransitions[key].erase(to);
}

template < class SymbolType, class StateType >
bool VisiblyPushdownNPDA < SymbolType, StateType >::removeLocalTransition(const StateType& from, const SymbolType& input, const StateType& to) {
	std::pair<StateType, SymbolType> key(from, input);

	return localTransitions[key].erase(to);
}

template < class SymbolType, class StateType >
const std::map<std::pair<StateType, SymbolType>, std::set<std::pair<StateType, SymbolType> > >& VisiblyPushdownNPDA < SymbolType, StateType >::getCallTransitions() const {
	return callTransitions;
}

template < class SymbolType, class StateType >
const std::map<std::tuple<StateType, SymbolType, SymbolType>, std::set<StateType> >& VisiblyPushdownNPDA < SymbolType, StateType >::getReturnTransitions() const {
	return returnTransitions;
}

template < class SymbolType, class StateType >
const std::map<std::pair<StateType, SymbolType>, std::set<StateType> >& VisiblyPushdownNPDA < SymbolType, StateType >::getLocalTransitions() const {
	return localTransitions;
}

template < class SymbolType, class StateType >
int VisiblyPushdownNPDA < SymbolType, StateType >::compare(const VisiblyPushdownNPDA& other) const {
	auto first = std::tie(getStates(), getCallInputAlphabet(), getReturnInputAlphabet(), getLocalInputAlphabet(), getInitialStates(), getFinalStates(), getPushdownStoreAlphabet(), getBottomOfTheStackSymbol(), callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.getStates(), other.getCallInputAlphabet(), other.getReturnInputAlphabet(), other.getLocalInputAlphabet(), other.getInitialStates(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getBottomOfTheStackSymbol(), other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType, class StateType >
void VisiblyPushdownNPDA < SymbolType, StateType >::operator>>(std::ostream& out) const {
	out << "(VisiblyPushdownNPDA "
		<< "states = " << getStates()
		<< "callAlphabet = " << getCallInputAlphabet()
		<< "returnAlphabet = " << getReturnInputAlphabet()
		<< "localAlphabet = " << getLocalInputAlphabet()
		<< "initialStates = " << getInitialStates()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "bottomOfTheStackSymbol = " << getBottomOfTheStackSymbol()
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

template < class SymbolType, class StateType >
VisiblyPushdownNPDA < SymbolType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class SymbolType, class StateType >
VisiblyPushdownNPDA < SymbolType, StateType > VisiblyPushdownNPDA < SymbolType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, VisiblyPushdownNPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<SymbolType> callInputAlphabet = AutomatonFromXMLParser::parseCallInputAlphabet<SymbolType>(input);
	std::set<SymbolType> returnInputAlphabet = AutomatonFromXMLParser::parseReturnInputAlphabet<SymbolType>(input);
	std::set<SymbolType> localInputAlphabet = AutomatonFromXMLParser::parseLocalInputAlphabet<SymbolType>(input);
	std::set<SymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<SymbolType>(input);
	std::set<StateType> initialStates = AutomatonFromXMLParser::parseInitialStates<StateType>(input);
	SymbolType bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol<SymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	VisiblyPushdownNPDA < SymbolType, StateType > automaton(std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setCallInputAlphabet(std::move(callInputAlphabet));
	automaton.setReturnInputAlphabet(std::move(returnInputAlphabet));
	automaton.setLocalInputAlphabet(std::move(localInputAlphabet));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setInitialStates(std::move(initialStates));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<VisiblyPushdownNPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, VisiblyPushdownNPDA::getXmlTagName());
	return automaton;
}

template < class SymbolType, class StateType >
void VisiblyPushdownNPDA < SymbolType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, VisiblyPushdownNPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
		SymbolType push = AutomatonFromXMLParser::parseTransitionSinglePush<SymbolType>(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
		SymbolType pop = AutomatonFromXMLParser::parseTransitionSinglePop<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<SymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

template < class SymbolType, class StateType >
void VisiblyPushdownNPDA < SymbolType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(VisiblyPushdownNPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeCallInputAlphabet(out, this->getCallInputAlphabet());
	AutomatonToXMLComposer::composeReturnInputAlphabet(out, this->getReturnInputAlphabet());
	AutomatonToXMLComposer::composeLocalInputAlphabet(out, this->getLocalInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialStates(out, this->getInitialStates());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(VisiblyPushdownNPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class StateType >
void VisiblyPushdownNPDA < SymbolType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionSinglePush(out, target.second);

			out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getReturnTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getLocalTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType, class StateType >
alib::ObjectBase* VisiblyPushdownNPDA < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, SymbolType, automaton::CallAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, SymbolType>, std::set < std::pair<StateType, SymbolType> > >& callTransition : automaton.getCallTransitions())
			if (symbol == callTransition.first.second)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getLocalInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in local alphabet");
		if(automaton.getReturnInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in return alphabet");
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, SymbolType, automaton::ReturnAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, SymbolType, SymbolType>, std::set < StateType> >& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<1>(returnTransition.first))
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getLocalInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in local alphabet");
		if(automaton.getCallInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in call alphabet");
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, SymbolType, automaton::LocalAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, SymbolType>, std::set < StateType >>& localTransition : automaton.getLocalTransitions())
			if (symbol == localTransition.first.second)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		if(automaton.getReturnInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in return alphabet");
		if(automaton.getCallInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + std::to_string ( symbol ) + " already in call alphabet");
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, SymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, SymbolType>, std::set<std::pair<StateType, SymbolType> > >& callTransition : automaton.getCallTransitions())
			for(const std::pair<StateType, SymbolType>& to : callTransition.second)
				if (symbol == to.second)
					return true;

		for (const std::pair<const std::tuple<StateType, SymbolType, SymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<2>(returnTransition.first))
				return true;

		if(automaton.getBottomOfTheStackSymbol() == symbol)
			return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class StateType >
class ElementConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, SymbolType, automaton::BottomOfTheStackSymbol > {
public:
	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialStates ( ).count ( state ) )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<StateType, SymbolType>, std::set<std::pair<StateType, SymbolType> > >& callTransition : automaton.getCallTransitions()) {
			if (state == callTransition.first.first)
				return true;
			for(const std::pair<StateType, SymbolType>& target : callTransition.second) {
				if(target.first == state)
					return true;
			}
		}

		for (const std::pair<const std::tuple<StateType, SymbolType, SymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions()) {
			if (state == std::get<0>(returnTransition.first))
				return true;
			for(const StateType& target : returnTransition.second) {
				if(target == state)
					return true;
			}
		}

		for (const std::pair<const std::pair<StateType, SymbolType>, std::set<StateType> >& localTransition : automaton.getLocalTransitions()) {
			if (state == localTransition.first.first)
				return true;
			for(const StateType& target : localTransition.second) {
				if(target == state)
					return true;
			}
		}

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::VisiblyPushdownNPDA < SymbolType, StateType >, StateType, automaton::InitialStates > {
public:
	static bool used ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::VisiblyPushdownNPDA < SymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* VISIBLY_PUSHDOWN_NPDA_H_ */
