/*
 * FormalRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_ELEMENT_H_
#define FORMAL_REG_EXP_ELEMENT_H_

#include <set>
#include <tree>
#include <core/visitor.hpp>
#include "../RegExpFeatures.h"

namespace regexp {

/**
 * Abstract class representing element in the formal regular expression. Can be operator or symbol.
 */
template < class SymbolType >
class FormalRegExpElement : public alib::CommonBase < FormalRegExpElement < SymbolType > >, public std::BaseNode < FormalRegExpElement < SymbolType > > {
public:
	class Visitor {
	public:
		virtual void visit ( const FormalRegExpAlternation < SymbolType > & ) = 0;
		virtual void visit ( const FormalRegExpConcatenation < SymbolType > & ) = 0;
		virtual void visit ( const FormalRegExpIteration < SymbolType > & ) = 0;
		virtual void visit ( const FormalRegExpSymbol < SymbolType > & ) = 0;
		virtual void visit ( const FormalRegExpEmpty < SymbolType > & ) = 0;
		virtual void visit ( const FormalRegExpEpsilon < SymbolType > & ) = 0;
	};

	template < class ReturnType, class Visitor, class ... Params >
	class VisitorContext : public std::VisitorContextAux < ReturnType, Visitor, Params ... >, public FormalRegExpElement::Visitor {
	public:
		using std::VisitorContextAux < ReturnType, Visitor, Params ... >::VisitorContextAux;

		void visit ( const FormalRegExpAlternation < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRegExpConcatenation < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRegExpIteration < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRegExpSymbol < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRegExpEmpty < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRegExpEpsilon < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}
	};

	template < class ReturnType, class Visitor, class ... Params >
	ReturnType accept ( Params && ... params ) const {
		VisitorContext < ReturnType, Visitor, Params ... > context ( std::forward < Params > ( params ) ... );
		accept ( context );
		return context.getResult ( );
	}

	virtual void accept ( FormalRegExpElement::Visitor & visitor ) const = 0;

	/**
	 * Creates copy of the element.
	 *
	 * @return copy of the element
	 */
	virtual std::smart_ptr < UnboundedRegExpElement < SymbolType > > asUnbounded ( ) const = 0;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const SymbolType & symbol ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const = 0;

	/**
	 * Traverses the regexp tree and checks whether all symbols in the regexp tree are in the alphabet
	 *
	 * @param alphabet
	 * @return true if symbols in the regexp are in the alphabet, false otherwise
	 */
	virtual bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @return the minimal alphabet needed by the regexp
	 */
	std::set < SymbolType > computeMinimalAlphabet ( ) const;
};

template < class SymbolType >
std::set < SymbolType > FormalRegExpElement < SymbolType >::computeMinimalAlphabet ( ) const {
	std::set < SymbolType > res;

	computeMinimalAlphabet ( res );
	return res;
}

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_ELEMENT_H_ */
