/*
 * FormalRegExpStructure.h
 *
 *  Created on: Oct 20, 2016
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_STRUCTURE_H_
#define FORMAL_REG_EXP_STRUCTURE_H_

#include <exception/CommonException.h>

#include "../RegExpFeatures.h"
#include "../common/RegExpFromXMLParser.h"
#include "../common/RegExpToXMLComposer.h"

#include "FormalRegExpEmpty.h"

namespace regexp {

/**
 * Represents formal regular expression structure as parsed from the XML. Regular expression is stored
 * as a tree of FormalRegExpElement.
 */
template < class SymbolType >
class FormalRegExpStructure {
protected:
	std::smart_ptr < FormalRegExpElement < SymbolType > > m_structure;

public:
	explicit FormalRegExpStructure ( );

	explicit FormalRegExpStructure ( FormalRegExpElement < SymbolType > && structure );
	explicit FormalRegExpStructure ( const FormalRegExpElement < SymbolType > & structure );
	explicit FormalRegExpStructure ( const UnboundedRegExpStructure < SymbolType > & other );

	/**
	 * @return Root node of the formal regular expression tree
	 */
	const FormalRegExpElement < SymbolType > & getStructure ( ) const;

	/**
	 * @return Root node of the formal regular expression tree
	 */
	FormalRegExpElement < SymbolType > & getStructure ( );

	/**
	 * Sets the root node of the formal regular expression tree
	 * @param structure root node to set
	 */
	void setStructure ( FormalRegExpElement < SymbolType > && structure );
	void setStructure ( const FormalRegExpElement < SymbolType > & structure );

	friend std::ostream & operator << ( std::ostream & out, const FormalRegExpStructure < SymbolType > & structure ) {
		out << structure.getStructure ( );
		return out;
	}

	friend bool operator < ( const FormalRegExpStructure < SymbolType > & first, const FormalRegExpStructure < SymbolType > & second ) {
		return first.getStructure().compare(second.getStructure()) < 0;
	}

	friend bool operator == ( const FormalRegExpStructure < SymbolType > & first, const FormalRegExpStructure < SymbolType > & second ) {
		return first.getStructure().compare(second.getStructure()) == 0;
	}
};

} /* namespace regexp */

namespace alib {

template < class SymbolType >
struct xmlApi < regexp::FormalRegExpStructure < SymbolType > > {
	static regexp::FormalRegExpStructure < SymbolType > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::FormalRegExpStructure < SymbolType > & data );
};

} /* namespace alib */

namespace std {

template < class SymbolType >
struct compare < regexp::FormalRegExpStructure < SymbolType > > {
	int operator()(const regexp::FormalRegExpStructure < SymbolType > & first, const regexp::FormalRegExpStructure < SymbolType > & second) const {
		return first.getStructure().compare(second.getStructure());
	}
};

} /* namespace std */

#include "../unbounded/UnboundedRegExpStructure.h"

namespace regexp {

template < class SymbolType >
FormalRegExpStructure < SymbolType >::FormalRegExpStructure ( FormalRegExpElement < SymbolType > && structure ) : m_structure ( NULL ) {
	setStructure ( std::move ( structure ) );
}

template < class SymbolType >
FormalRegExpStructure < SymbolType >::FormalRegExpStructure ( const FormalRegExpElement < SymbolType > & structure ) : FormalRegExpStructure ( std::move_copy ( structure ) ) {
}

template < class SymbolType >
FormalRegExpStructure < SymbolType >::FormalRegExpStructure ( ) : FormalRegExpStructure ( FormalRegExpEmpty < SymbolType > ( ) ) {
}

template < class SymbolType >
FormalRegExpStructure < SymbolType >::FormalRegExpStructure ( const UnboundedRegExpStructure < SymbolType > & other ) : FormalRegExpStructure ( * other.getStructure ( ).asFormal ( ) ) {
}

template < class SymbolType >
const FormalRegExpElement < SymbolType > & FormalRegExpStructure < SymbolType >::getStructure ( ) const {
	return * m_structure;
}

template < class SymbolType >
FormalRegExpElement < SymbolType > & FormalRegExpStructure < SymbolType >::getStructure ( ) {
	return * m_structure;
}

template < class SymbolType >
void FormalRegExpStructure < SymbolType >::setStructure ( const FormalRegExpElement < SymbolType > & structure ) {
	setStructure ( std::move_copy ( structure ) );
}

template < class SymbolType >
void FormalRegExpStructure < SymbolType >::setStructure ( FormalRegExpElement < SymbolType > && param ) {
	this->m_structure = std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( param ).plunder ( ) );
}

} /* namespace regexp */

namespace alib {

template < class SymbolType >
regexp::FormalRegExpStructure < SymbolType > xmlApi < regexp::FormalRegExpStructure < SymbolType > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return regexp::FormalRegExpStructure < SymbolType > ( regexp::RegExpFromXMLParser::parseFormalRegExpElement < SymbolType > ( input ) );
}

template < class SymbolType >
bool xmlApi < regexp::FormalRegExpStructure < SymbolType > >::first ( const std::deque < sax::Token >::const_iterator & ) {
	throw exception::CommonException ( "Unimplemented" );
}

template < class SymbolType >
void xmlApi < regexp::FormalRegExpStructure < SymbolType > >::compose ( std::deque < sax::Token > & output, const regexp::FormalRegExpStructure < SymbolType > & data ) {
	data.getStructure ( ).template accept < void, regexp::RegExpToXMLComposer::Formal > ( output );
}

} /* namespace alib */

#endif /* FORMAL_REG_EXP_STRUCTURE_H_ */
