/*
 * FormalRTEElement.h
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_ELEMENT_H_
#define FORMAL_RTE_ELEMENT_H_

#include <alphabet/ranked_symbol.hpp>
#include <set>
#include <tree>
#include <core/visitor.hpp>
#include "../RTEFeatures.h"

namespace rte {

template < class SymbolType, class RankType >
class FormalRTEElement : public alib::CommonBase < FormalRTEElement < SymbolType, RankType > >, public std::BaseNode < FormalRTEElement < SymbolType, RankType > > {
public:
	class Visitor {
	public:
		virtual void visit ( const FormalRTEAlternation < SymbolType, RankType > & alternation ) = 0;
		virtual void visit ( const FormalRTEIteration < SymbolType, RankType > & iteration ) = 0;
		virtual void visit ( const FormalRTESubstitution < SymbolType, RankType > & substitution ) = 0;
		virtual void visit ( const FormalRTESymbolAlphabet < SymbolType, RankType > & symbol ) = 0;
		virtual void visit ( const FormalRTESymbolSubst < SymbolType, RankType > & symbol ) = 0;
		virtual void visit ( const FormalRTEEmpty < SymbolType, RankType > & empty ) = 0;
	};

	template < class ReturnType, class Visitorr, class ... Params >
	class VisitorContext : public std::VisitorContextAux < ReturnType, Visitorr, Params ... >, public FormalRTEElement::Visitor {
	public:
		using std::VisitorContextAux < ReturnType, Visitorr, Params ... >::VisitorContextAux;

		void visit ( const FormalRTEAlternation < SymbolType, RankType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRTEIteration < SymbolType, RankType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRTESubstitution < SymbolType, RankType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRTESymbolAlphabet < SymbolType, RankType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRTESymbolSubst < SymbolType, RankType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const FormalRTEEmpty < SymbolType, RankType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}
	};

	template < class ReturnType, class Visitorr, class ... Params >
	ReturnType accept ( Params && ... params ) const {
		VisitorContext < ReturnType, Visitorr, Params ... > context ( std::forward < Params > ( params ) ... );
		accept ( context );
		return context.getResult ( );
	}

	virtual void accept ( FormalRTEElement::Visitor & visitor ) const = 0;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const = 0;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	std::pair < std::set < std::ranked_symbol < SymbolType, RankType > >, std::set < std::ranked_symbol < SymbolType, RankType > > > computeMinimalAlphabets ( ) const;

	/**
	 * Traverses the regexp tree and checks whether all symbols in the regexp tree are in the alphabet
	 *
	 * @param alphabet
	 * @return true if symbols in the regexp are in the alphabet, false otherwise
	 */
	virtual bool checkAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const = 0;
};

template < class SymbolType, class RankType >
std::pair < std::set < std::ranked_symbol < SymbolType, RankType > >, std::set < std::ranked_symbol < SymbolType, RankType > > > FormalRTEElement < SymbolType, RankType >::computeMinimalAlphabets ( ) const {
	std::set < std::ranked_symbol < SymbolType, RankType > > alphabetF;
	std::set < std::ranked_symbol < SymbolType, RankType > > alphabetK;

	computeMinimalAlphabet ( alphabetF, alphabetK );
	return std::make_pair ( std::move ( alphabetF ), std::move ( alphabetK ) );
}

} /* namespace rte */

#endif /* FORMAL_RTE_ELEMENT_H_ */
