/*
 * FormalRTESubstitution.h
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_SUBSTITUTION_H_
#define FORMAL_RTE_SUBSTITUTION_H_

#include "FormalRTEElement.h"
#include "FormalRTESymbolSubst.h"

#include <exception/CommonException.h>
#include <sstream>

namespace rte {

/**
 * Represents concatenation operator in the regular expression. Contains list of FormalRTEElement
 * as operands of the operator.
 */
template < class SymbolType, class RankType >
class FormalRTESubstitution : public FormalRTEElement < SymbolType, RankType >, public std::BinaryNode < std::smart_ptr < FormalRTEElement < SymbolType, RankType > >, std::smart_ptr < FormalRTEElement < SymbolType, RankType > >, FormalRTESubstitution < SymbolType, RankType > > {
protected:
	std::smart_ptr < FormalRTESymbolSubst < SymbolType, RankType > > m_substitutionSymbol; // substite this in left by right

public:
	void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const {
		visitor.visit ( * this );
	}

	explicit FormalRTESubstitution ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol );
	explicit FormalRTESubstitution ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol );

	/**
	 * @copydoc FormalRTEElement::clone() const
	 */
	virtual FormalRTESubstitution < SymbolType, RankType > * clone ( ) const;

	/**
	 * @copydoc FormalRTEElement::plunder() const
	 */
	virtual FormalRTESubstitution < SymbolType, RankType > * plunder ( ) &&;

	/**
	 * @copydoc FormalRTEElement::testSymbol() const
	 */
	virtual bool testSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) const;

	/**
	 * @copydoc FormalRTEElement::computeMinimalAlphabet()
	 */
	virtual void computeMinimalAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const;

	/**
	 * @copydoc FormalRTEElement::checkAlphabet()
	 */
	virtual bool checkAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const;

	/**
	 * @return elements
	 */
	const FormalRTEElement < SymbolType, RankType > & getLeftElement ( ) const;
	const FormalRTEElement < SymbolType, RankType > & getRightElement ( ) const;
	const FormalRTESymbolSubst < SymbolType, RankType > & getSubstitutionSymbol ( ) const;

	/**
	 * @return elements
	 */
	FormalRTEElement < SymbolType, RankType > & getLeftElement ( );
	FormalRTEElement < SymbolType, RankType > & getRightElement ( );
	FormalRTESymbolSubst < SymbolType, RankType > & getSubstitutionSymbol ( );

	/**
	 * @param element to append
	 */
	void setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element );
	void setLeftElement ( FormalRTEElement < SymbolType, RankType > && element );

	void setRightElement ( const FormalRTEElement < SymbolType, RankType > & element );
	void setRightElement ( FormalRTEElement < SymbolType, RankType > && element );

	void setSubstitutionSymbol ( FormalRTESymbolSubst < SymbolType, RankType > element );

	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FormalRTESubstitution & ) const;

	/**
	 * @copydoc FormalRTEElement::operator>>() const
	 */
	virtual void operator >>( std::ostream & out ) const;
	virtual explicit operator std::string ( ) const;
};

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType >::FormalRTESubstitution ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol ) : std::BinaryNode < std::smart_ptr < FormalRTEElement < SymbolType, RankType > >, std::smart_ptr < FormalRTEElement < SymbolType, RankType > >, FormalRTESubstitution < SymbolType, RankType > > ( std::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( left ).plunder ( ) ), std::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( right ).plunder ( ) ) ), m_substitutionSymbol ( std::smart_ptr < FormalRTESymbolSubst < SymbolType, RankType > > ( std::move ( substitutionSymbol ).plunder ( ) ) ) {
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType >::FormalRTESubstitution ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol ) : FormalRTESubstitution ( std::move_copy ( left ), std::move_copy ( right ), std::move ( substitutionSymbol ) ) {
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getLeftElement ( ) const {
	return * this->getLeft ( );
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getRightElement ( ) const {
	return * this->getRight ( );
}

template < class SymbolType, class RankType >
const FormalRTESymbolSubst < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getSubstitutionSymbol ( ) const {
	return * this->m_substitutionSymbol;
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getLeftElement ( ) {
	return * this->getLeft ( );
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getRightElement ( ) {
	return * this->getRight ( );
}

template < class SymbolType, class RankType >
FormalRTESymbolSubst < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getSubstitutionSymbol ( ) {
	return * this->m_substitutionSymbol;
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setLeftElement ( std::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setLeftElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setLeft ( std::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setRightElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setRightElement ( std::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setRightElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setRight ( std::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setSubstitutionSymbol ( FormalRTESymbolSubst < SymbolType, RankType > symbol ) {
	m_substitutionSymbol = std::smart_ptr < FormalRTESymbolSubst < SymbolType, RankType > > ( std::move ( symbol ).plunder ( ) );
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType > * FormalRTESubstitution < SymbolType, RankType >::clone ( ) const {
	return new FormalRTESubstitution ( * this );
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType > * FormalRTESubstitution < SymbolType, RankType >::plunder ( ) && {
	return new FormalRTESubstitution ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
int FormalRTESubstitution < SymbolType, RankType >::compare ( const FormalRTESubstitution < SymbolType, RankType > & other ) const {
	auto first = std::tie ( getLeftElement ( ), getRightElement ( ), getSubstitutionSymbol ( ) );
	auto second = std::tie ( other.getLeftElement ( ), other.getRightElement ( ), other.getSubstitutionSymbol ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTESubstitution";
	out << " " << getLeftElement ( );
	out << " " << * m_substitutionSymbol;
	out << " " << getRightElement ( );
	out << ")";
}

template < class SymbolType, class RankType >
bool FormalRTESubstitution < SymbolType, RankType >::testSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) const {
	if ( getLeftElement ( ).testSymbol ( symbol ) ) return true;

	if ( getRightElement ( ).testSymbol ( symbol ) ) return true;

	if ( m_substitutionSymbol->getSymbol ( ) == symbol ) return true;

	return false;
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::computeMinimalAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	alphabetK.insert ( m_substitutionSymbol->getSymbol ( ) );
	getLeftElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
	getRightElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
bool FormalRTESubstitution < SymbolType, RankType >::checkAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetF, const std::set < std::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return alphabetK.count ( getSubstitutionSymbol ( ).getSymbol ( ) ) > 0 && getLeftElement ( ).checkAlphabet ( alphabetF, alphabetK ) && getRightElement ( ).checkAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_SUBSTITUTION_H_ */
