/*
 * RankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_TREE_H_
#define RANKED_TREE_H_

#include <string>
#include <set>
#include <tree>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include "alphabet/ranked_symbol.hpp"

#include "../TreeFeatures.h"
#include "../RankedTreeBase.h"
#include "../TreeException.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"

namespace tree {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType, class RankType >
class RankedTree : public RankedTreeBase, public std::Components < RankedTree < SymbolType, RankType >, std::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::tree < std::ranked_symbol < SymbolType, RankType > > content;

	void checkAlphabet ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & pattern ) const;

	void checkArities ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & data ) const;

public:
	/**
	 * @copydoc std::tree < std::ranked_symbol < SymbolType, RankType > >::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc std::tree < std::ranked_symbol < SymbolType, RankType > >::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedTree ( std::set < std::ranked_symbol < SymbolType, RankType > > alphabet, std::tree < std::ranked_symbol < SymbolType, RankType > > tree );
	explicit RankedTree ( std::tree < std::ranked_symbol < SymbolType, RankType > > tree );
	explicit RankedTree ( const UnrankedTree < SymbolType > & other );

	/**
	 * @return Root node of the regular expression tree
	 */
	const std::tree < std::ranked_symbol < SymbolType, RankType > > & getContent ( ) const;

	const std::set < std::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( std::tree < std::ranked_symbol < SymbolType, RankType > > tree );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RankedTree";

		return xmlTagName;
	}

	static RankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

} /* namespace tree */

#include "../unranked/UnrankedTree.h"

namespace tree {

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( std::set < std::ranked_symbol < SymbolType, RankType > > alphabet, std::tree < std::ranked_symbol < SymbolType, RankType > > tree ) : std::Components < RankedTree, std::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), content ( std::move ( tree ) ) {
	checkAlphabet ( content );
	checkArities ( content );
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( std::tree < std::ranked_symbol < SymbolType, RankType > > tree ) : RankedTree ( std::set < std::ranked_symbol < SymbolType, RankType > > ( tree.prefix_begin ( ), tree.prefix_end ( ) ), tree ) {
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( const UnrankedTree < SymbolType > & other ) : RankedTree ( TreeAuxiliary::toRankedTree < SymbolType, RankType > ( other.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
RankedTreeBase * RankedTree < SymbolType, RankType >::clone ( ) const {
	return new RankedTree ( * this );
}

template < class SymbolType, class RankType >
RankedTreeBase * RankedTree < SymbolType, RankType >::plunder ( ) && {
	return new RankedTree ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const std::tree < std::ranked_symbol < SymbolType, RankType > > & RankedTree < SymbolType, RankType >::getContent ( ) const {
	return content;
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::checkAlphabet ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & data ) const {
	std::set < std::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < std::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::checkArities ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & data ) const {
	if ( ( size_t ) data.getData ( ).getRank ( ) != data.getChildren ( ).size ( ) )
		throw exception::CommonException ( "Invalid rank." );

	for ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & child : data )
		checkArities ( child );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::setTree ( std::tree < std::ranked_symbol < SymbolType, RankType > > tree ) {
	checkAlphabet ( tree );
	checkArities ( tree );

	this->content = std::move ( tree );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(RankedTree " << this->content << ")";
}

template < class SymbolType, class RankType >
int RankedTree < SymbolType, RankType >::compare ( const RankedTree & other ) const {
	auto first = std::tie ( content, getAlphabet() );
	auto second = std::tie ( other.content, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType > RankedTree < SymbolType, RankType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedTree::getXmlTagName() );
	std::set < std::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < std::ranked_symbol < SymbolType, RankType > > ( input );
	std::tree < std::ranked_symbol < SymbolType, RankType > > root = TreeFromXMLParser::parseTreeContent < std::ranked_symbol < SymbolType, RankType > > ( input );
	RankedTree < SymbolType, RankType > tree ( std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedTree::getXmlTagName() );
	return tree;
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( RankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
alib::ObjectBase* RankedTree < SymbolType, RankType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace std {

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::RankedTree < SymbolType, RankType >, std::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::RankedTree < SymbolType, RankType > & tree, const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		const std::tree < std::ranked_symbol < SymbolType, RankType > > & content = tree.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end();
	}

	static bool available ( const ::tree::RankedTree < SymbolType, RankType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::RankedTree < SymbolType, RankType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
	}
};

} /* namespace std */

#endif /* RANKED_TREE_H_ */
