/*
 * UnrankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef UNRANKED_TREE_H_
#define UNRANKED_TREE_H_

#include <string>
#include <set>
#include <tree>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "../TreeFeatures.h"
#include "../UnrankedTreeBase.h"
#include "../TreeException.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"

namespace tree {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType >
class UnrankedTree : public UnrankedTreeBase, public std::Components < UnrankedTree < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::tree < SymbolType > content;

	void checkAlphabet ( const std::tree < SymbolType > & pattern ) const;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedTree ( std::set < SymbolType > alphabet, std::tree < SymbolType > tree );
	explicit UnrankedTree ( std::tree < SymbolType > tree );
	template < class RankType >
	explicit UnrankedTree ( const RankedTree < SymbolType, RankType > & other );

	const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * @return content of the tree
	 */
	const std::tree < SymbolType > & getContent ( ) const;

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( std::tree < SymbolType > data );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedTree";

		return xmlTagName;
	}

	static UnrankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;
};

} /* namespace tree */

#include "../ranked/RankedTree.h"

namespace tree {

template < class SymbolType >
UnrankedTree < SymbolType >::UnrankedTree ( std::set < SymbolType > alphabet, std::tree < SymbolType > tree ) : std::Components < UnrankedTree, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), content ( std::move ( tree ) ) {
	checkAlphabet ( content );
}

template < class SymbolType >
UnrankedTree < SymbolType >::UnrankedTree ( std::tree < SymbolType > pattern ) : UnrankedTree ( std::set < SymbolType > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ), pattern ) {
}

template < class SymbolType >
template < class RankType >
UnrankedTree < SymbolType >::UnrankedTree ( const RankedTree < SymbolType, RankType > & other ) : UnrankedTree ( TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), TreeAuxiliary::toUnrankedTree ( other.getContent ( ) ) ) {
}

template < class SymbolType >
UnrankedTreeBase * UnrankedTree < SymbolType >::clone ( ) const {
	return new UnrankedTree ( * this );
}

template < class SymbolType >
UnrankedTreeBase * UnrankedTree < SymbolType >::plunder ( ) && {
	return new UnrankedTree ( std::move ( * this ) );
}

template < class SymbolType >
const std::tree < SymbolType > & UnrankedTree < SymbolType >::getContent ( ) const {
	return content;
}

template < class SymbolType >
void UnrankedTree < SymbolType >::checkAlphabet ( const std::tree < SymbolType > & data ) const {
	std::set < SymbolType > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < SymbolType > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::setTree ( std::tree < SymbolType > data ) {
	checkAlphabet ( data );

	this->content = std::move ( data );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(UnrankedTree " << this->content << ")";
}

template < class SymbolType >
int UnrankedTree < SymbolType >::compare ( const UnrankedTree & other ) const {
	auto first = std::tie ( content, getAlphabet() );
	auto second = std::tie ( other.content, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void UnrankedTree < SymbolType >::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

template < class SymbolType >
UnrankedTree < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
UnrankedTree < SymbolType > UnrankedTree < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedTree::getXmlTagName() );
	std::set < SymbolType > rankedAlphabet = TreeFromXMLParser::parseAlphabet < SymbolType > ( input );
	std::tree < SymbolType > root = TreeFromXMLParser::parseTreeContent < SymbolType > ( input );
	UnrankedTree < SymbolType > tree ( std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedTree::getXmlTagName() );
	return tree;
}

template < class SymbolType >
void UnrankedTree < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnrankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( UnrankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* UnrankedTree < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace std {

template < class SymbolType >
class ComponentConstraint< ::tree::UnrankedTree < SymbolType >, SymbolType, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::UnrankedTree < SymbolType > & tree, const SymbolType & symbol ) {
		const std::tree<SymbolType>& content = tree.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end();
	}

	static bool available ( const ::tree::UnrankedTree < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::tree::UnrankedTree < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* UNRANKED_TREE_H_ */
