/*
 * AdjacencyListDirectedGraph.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "AdjacencyListDirectedGraph.h"

#include <compare>

#include <object/Object.h>
#include <core/xmlApi.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
//#include "utils.h"


namespace graph {

template<typename T>
inline void listRemoveOne(std::list<T> &lst, const T &item)
{
	lst.erase(std::find(lst.begin(), lst.end(), item));
}

AdjacencyListDirectedGraph::AdjacencyListDirectedGraph() : DirectedGraph() {
}

AdjacencyListDirectedGraph::AdjacencyListDirectedGraph( const DirectedGraph & graph )
{
	DirectedGraph::fromDirected( graph );
}

AdjacencyListDirectedGraph::AdjacencyListDirectedGraph( const UndirectedGraph & ugraph )
{
	DirectedGraph::fromUndirected( ugraph );
}

GraphBase * AdjacencyListDirectedGraph::clone() const
{
	return new AdjacencyListDirectedGraph( *this );
}

GraphBase * AdjacencyListDirectedGraph::plunder() &&
{
	return new AdjacencyListDirectedGraph( std::move( *this ) );
}

REPRESENTATION AdjacencyListDirectedGraph::getRepresentation() const
{
	return REPRESENTATION::ADJACENCY_LIST;
}

std::string AdjacencyListDirectedGraph::className() const
{
	return "AdjacencyListDirectedGraph";
}

std::set<Node> AdjacencyListDirectedGraph::getNodes() const
{
	std::set<Node> out;

	for (auto i : succ) {
		out.insert(i.first);
	}

	return out;
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::getEdges() const
{
	return edges;
}

size_t AdjacencyListDirectedGraph::nodeCount() const
{
	return succ.size();
}

size_t AdjacencyListDirectedGraph::edgeCount() const
{
	return edges.size();
}

bool AdjacencyListDirectedGraph::hasNode( const Node & node ) const
{
	return succ.find(node) != succ.end();
}

bool AdjacencyListDirectedGraph::hasEdge( const DirectedEdge & edge ) const
{
	return edges.find(edge) != edges.end();
}

bool AdjacencyListDirectedGraph::hasEdge( const Node & from, const Node & to ) const
{
	auto search = succ.find(from);
	if (search == succ.end()) {
		return false;
	}
	return std::find(search->second.begin(), search->second.end(), to) != search->second.end();
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::findEdges( const Node & from, const Node & to ) const
{
	std::set<DirectedEdge> out;

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == from && e.getToNode() == to) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyListDirectedGraph::neighbors( const Node & node ) const
{
	std::set<Node> out;

	auto search = succ.find(node);
	if (search == succ.end()) {
		return out;
	}

	std::copy(search->second.begin(), search->second.end(), std::inserter(out, out.begin()));
	return out;
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::neighborEdges( const Node & node ) const
{
	std::set<DirectedEdge> out;
	const std::set<Node> &nghbrs = neighbors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyListDirectedGraph::successors( const Node & node ) const
{
	std::set<Node> out;

	auto search = succ.find(node);
	if (search == succ.end()) {
		return out;
	}

	std::copy(search->second.begin(), search->second.end(), std::inserter(out, out.begin()));
	return out;
}

std::set<Node> AdjacencyListDirectedGraph::predecessors( const Node & node ) const
{
	std::set<Node> out;

	auto search = pred.find(node);
	if (search == pred.end()) {
		return out;
	}

	std::copy(search->second.begin(), search->second.end(), std::inserter(out, out.begin()));
	return out;
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::successorEdges( const Node & node ) const
{
	std::set<DirectedEdge> out;
	const std::set<Node> &nghbrs = successors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

std::set<DirectedEdge> AdjacencyListDirectedGraph::predecessorEdges( const Node & node ) const
{
	std::set<DirectedEdge> out;
	const std::set<Node> &nghbrs = predecessors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyListDirectedGraph::addNode( const Node & node )
{
	if (succ.find(node) != succ.end())
		return false;

	succ[node]; // create empty list for this key
	pred[node];
	return true;
}

bool AdjacencyListDirectedGraph::addNode( const Node & node, int value )
{
	if (!addNode( node ))
		return false;

	DirectedGraph::addNode( node, value );
	return true;
}

bool AdjacencyListDirectedGraph::removeNode( const Node & node )
{
	auto search = succ.find(node);
	if (search == succ.end()) {
		return false;
	}
	succ.erase(search);

	search = pred.find(node);
	if (search == pred.end()) {
		return false; // should never happen
	}
	pred.erase(search);

	DirectedGraph::removeNode(node);

	for (auto& i : edges) {
		if (node == i.getFromNode()) {
			listRemoveOne(pred[i.getToNode()], node);
			edges.erase(i);
		} else if (node == i.getToNode()) {
			listRemoveOne(succ[i.getFromNode()], node);
			edges.erase(i);
		}
	}

	return true;
}

bool AdjacencyListDirectedGraph::addEdge( const DirectedEdge & edge )
{
	if (edges.find(edge) != edges.end())
		return false;
	if (succ.find(edge.getFromNode()) == succ.end())
		return false;
	if (succ.find(edge.getToNode()) == pred.end())
		return false;

	edges.insert(edge);

	succ[edge.getFromNode()].push_back(edge.getToNode());
	pred[edge.getToNode()].push_back(edge.getFromNode());

	return true;
}

bool AdjacencyListDirectedGraph::addEdge( const DirectedEdge & edge, int value )
{
    if (!addEdge( edge ))
		return false;

	DirectedGraph::addEdge( edge, value );
	return true;
}

bool AdjacencyListDirectedGraph::removeEdge( const DirectedEdge & edge )
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);
	DirectedGraph::removeEdge(edge);

	listRemoveOne(succ[edge.getFromNode()], edge.getToNode());
	listRemoveOne(pred[edge.getToNode()], edge.getFromNode());

	return true;
}

AdjacencyListDirectedGraph AdjacencyListDirectedGraph::parse( std::deque<sax::Token>::iterator & input )
{
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, getXmlTagName());

	AdjacencyListDirectedGraph graph;
	GraphFromXMLParser::parseNodes(input, graph);
	GraphFromXMLParser::parseDirectedEdges(input, graph);
	GraphFromXMLParser::parseNodeValues(input, graph);
	GraphFromXMLParser::parseDirectedEdgeValues(input, graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, getXmlTagName());
	return graph;
}

void AdjacencyListDirectedGraph::compose( std::deque<sax::Token> & out ) const
{
	out.emplace_back(getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

alib::ObjectBase* AdjacencyListDirectedGraph::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} // namespace graph

namespace alib {

auto adjListDirGraphParserRegister  = xmlApi<graph::Graph>::ParserRegister<graph::AdjacencyListDirectedGraph>();
auto adjListDirGraphParserRegister2 = xmlApi<alib::Object>::ParserRegister<graph::AdjacencyListDirectedGraph>();

} /* namespace alib */
