/*
 * AdjacencyListDirectedGraph.h
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#ifndef ADJACENCY_LIST_DIRECTED_GRAPH_H_
#define ADJACENCY_LIST_DIRECTED_GRAPH_H_

#include <list>
#include <unordered_map>

#include "DirectedGraph.h"

namespace graph {

class UndirectedGraph;

/*
 *  AdjacencyList representation
 *
 *  unordered_map:
 *    [ n1 ] -> list [ n2, n3, n4 ]
 *    [ n2 ] -> list [ n1, n1 ]      < multi-edge
 *    [ n3 ] -> list [ n4 ]
 *    [ n4 ] -> list [ n1, n2 ]
 *
 *  edges:
 *    n1 -> n2, n1 -> n3, n1 -> n4
 *    n2 -> n1, n2 -> n1
 *    ...
 */

class AdjacencyListDirectedGraph : public DirectedGraph {
	std::unordered_map< Node, std::list< Node > > succ;
	std::unordered_map< Node, std::list< Node > > pred;
	std::set< DirectedEdge > edges;

public:
	AdjacencyListDirectedGraph();

	/// constructs an directed adjacency list from any representation of directed graph
	AdjacencyListDirectedGraph( const DirectedGraph & graph );
	/// converts undirected graph into directed by making 2 directed edges in both directions from each undirected one
	AdjacencyListDirectedGraph( const UndirectedGraph & ugraph );

	virtual GraphBase * clone() const override;
	virtual GraphBase * plunder() && override;

	virtual REPRESENTATION getRepresentation() const override;
	virtual std::string className() const override;

	virtual std::set< Node > getNodes() const override;
	virtual std::set< DirectedEdge > getEdges() const override;

	virtual size_t nodeCount() const override;
	virtual size_t edgeCount() const override;

	virtual bool hasNode( const Node & node ) const override;
	virtual bool hasEdge( const DirectedEdge & edge ) const override;

	virtual bool hasEdge( const Node & from, const Node & to) const override;
	virtual std::set< DirectedEdge > findEdges( const Node & from, const Node & to) const override;

	virtual std::set< Node > neighbors( const Node & node ) const override;
	virtual std::set< DirectedEdge > neighborEdges( const Node & node ) const override;

	virtual std::set< Node > successors( const Node & node ) const override;
	virtual std::set< Node > predecessors( const Node & node ) const override;
	virtual std::set< DirectedEdge > successorEdges( const Node & node ) const override;
	virtual std::set< DirectedEdge > predecessorEdges( const Node & node ) const override;

	virtual bool addNode( const Node & node ) override;
	virtual bool addNode( const Node & node, int value ) override;
	virtual bool removeNode( const Node & node ) override;

	virtual bool addEdge( const DirectedEdge & edge ) override;
	virtual bool addEdge( const DirectedEdge & edge, int value ) override;
	virtual bool removeEdge( const DirectedEdge & edge ) override;

	static AdjacencyListDirectedGraph parse( std::deque< sax::Token >::iterator & input );
	virtual void compose( std::deque< sax::Token > & out ) const override;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "AdjacencyListDirectedGraph";

		return xmlTagName;
	}

	alib::ObjectBase * inc ( ) && override;
};

} // namespace graph

#endif // ADJACENCY_LIST_DIRECTED_GRAPH_H_
