/*
 * SuffixTrieNodeFinalMark.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "SuffixTrieNodeFinalMark.h"
#include "SuffixTrieNodeTerminatingSymbol.h"
#include "SuffixTrieFinalMark.h"
#include <exception/CommonException.h>
#include <sstream>

namespace indexes {

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( std::map < DefaultSymbolType, SuffixTrieNodeFinalMark * > children, bool finalMark ) : m_children ( std::move ( children ) ), m_finalMark ( finalMark ), parentTree ( NULL ) {
	for ( auto & element : this->m_children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( const SuffixTrieNodeTerminatingSymbol & other, const DefaultSymbolType & terminatingSymbol ) : m_finalMark ( false ), parentTree ( NULL ) {
	for ( auto & element : other.getChildren ( ) )
		if ( element.first == terminatingSymbol )
			this->m_finalMark = true;
		else
			this->m_children.insert ( std::make_pair ( element.first, new SuffixTrieNodeFinalMark ( * element.second, terminatingSymbol ) ) );

	for ( auto & element : this->m_children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( const SuffixTrieNodeFinalMark & other ) : m_finalMark ( other.m_finalMark ), parentTree ( NULL ) {
	for ( const auto & element : other.m_children )
		m_children.insert ( std::make_pair ( element.first, element.second->clone ( ) ) );

	for ( auto & element : this->m_children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( SuffixTrieNodeFinalMark && other ) noexcept : m_children ( std::move ( other.m_children ) ), m_finalMark ( other.m_finalMark ), parentTree ( NULL ) {
	other.m_children.clear ( );

	for ( auto & element : this->m_children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::operator =( const SuffixTrieNodeFinalMark & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieNodeFinalMark ( other );

	return * this;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::operator =( SuffixTrieNodeFinalMark && other ) noexcept {
	std::swap ( this->m_children, other.m_children );
	std::swap ( this->m_finalMark, other.m_finalMark );
	std::swap ( this->parentTree, other.parentTree ); // this->parentTree is stored within other.parentTree and it is reattached on the next line

	for ( auto & element : this->m_children )
		element.second->parent = this;

	this->attachTree ( other.parentTree );

	return * this;
}

SuffixTrieNodeFinalMark::~SuffixTrieNodeFinalMark ( ) noexcept {
	for ( auto element : m_children )
		delete element.second;

	m_children.clear ( );
}

const std::map < const DefaultSymbolType, const SuffixTrieNodeFinalMark * > & SuffixTrieNodeFinalMark::getChildren ( ) const {
	return * reinterpret_cast < const std::map < const DefaultSymbolType, const SuffixTrieNodeFinalMark * > * > ( & m_children );
}

bool SuffixTrieNodeFinalMark::getFinalMark ( ) const {
	return m_finalMark;
}

void SuffixTrieNodeFinalMark::setFinalMark ( bool newFinalMark ) {
	m_finalMark = newFinalMark;
}

const std::map < DefaultSymbolType, SuffixTrieNodeFinalMark * > & SuffixTrieNodeFinalMark::getChildren ( ) {
	return m_children;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::getChild ( const DefaultSymbolType & symbol ) {
	std::map < DefaultSymbolType, SuffixTrieNodeFinalMark * >::const_iterator iter = m_children.find ( symbol );

	if ( iter == m_children.end ( ) ) throw exception::CommonException ( "child does not exist" );

	return * iter->second;
}

const SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::getChild ( const DefaultSymbolType & symbol ) const {
	std::map < DefaultSymbolType, SuffixTrieNodeFinalMark * >::const_iterator iter = m_children.find ( symbol );

	if ( iter == m_children.end ( ) ) throw exception::CommonException ( "child does not exist" );

	return * iter->second;
}

bool SuffixTrieNodeFinalMark::hasChild ( const DefaultSymbolType & symbol ) const {
	if ( m_children.find ( symbol ) == m_children.end ( ) ) return false;

	return true;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::addChild ( DefaultSymbolType symbol, SuffixTrieNodeFinalMark node ) {
	std::map < DefaultSymbolType, SuffixTrieNodeFinalMark * >::iterator iter = m_children.find ( symbol );

	if ( iter != m_children.end ( ) ) throw exception::CommonException ( "child already exist" );

	if ( ( this->parentTree != NULL ) && ! ( this->parentTree->getAlphabet ( ).count ( symbol ) ) ) throw exception::CommonException ( "Symbol is not in the alphabet" );

	return * m_children.insert ( std::make_pair ( std::move ( symbol ), new SuffixTrieNodeFinalMark ( std::move ( node ) ) ) ).first->second;
}

SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::getParent ( ) {
	return parent;
}

const SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::getParent ( ) const {
	return parent;
}

void SuffixTrieNodeFinalMark::swap ( SuffixTrieNodeFinalMark & other ) {
	const SuffixTrieFinalMark * thisParentTree = this->parentTree;
	const SuffixTrieFinalMark * otherParentTree = other.parentTree;

	SuffixTrieNodeFinalMark tmp = std::move ( other );

	other  = std::move ( * this );
	* this = std::move ( tmp );

	this->attachTree ( thisParentTree );
	other.attachTree ( otherParentTree );
}

SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::clone ( ) const {
	return new SuffixTrieNodeFinalMark ( * this );
}

SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::plunder ( ) && {
	return new SuffixTrieNodeFinalMark ( std::move ( * this ) );
}

int SuffixTrieNodeFinalMark::compare ( const SuffixTrieNodeFinalMark & other ) const {
	auto first = std::tie ( getChildren() );
	auto second = std::tie ( other.getChildren() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void SuffixTrieNodeFinalMark::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieNodeFinalMark " << " children = " << this->m_children << " finalMark = " << this->m_finalMark << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieNodeFinalMark & node ) {
	node >> out;
	return out;
}

bool SuffixTrieNodeFinalMark::testSymbol ( const DefaultSymbolType & symbol ) const {
	for ( const auto & child : this->m_children ) {
		if ( symbol == child.first ) return true;

		if ( child.second->testSymbol ( symbol ) ) return true;
	}

	return false;
}

bool SuffixTrieNodeFinalMark::attachTree ( const SuffixTrieFinalMark * tree ) {
	if ( this->parentTree == tree ) return true;

	this->parentTree = tree;

	for ( const auto & child : this->m_children ) {
		if ( ( this->parentTree != NULL ) && ! ( this->parentTree->getAlphabet ( ).count ( child.first ) ) ) return false;

		if ( !child.second->attachTree ( tree ) ) return false;
	}

	return true;
}

std::set < DefaultSymbolType > SuffixTrieNodeFinalMark::computeMinimalAlphabet ( ) const {
	std::set < DefaultSymbolType > res;
	computeMinimalAlphabet ( res );
	return res;
}

void SuffixTrieNodeFinalMark::computeMinimalAlphabet ( std::set < DefaultSymbolType > & alphabet ) const {
	for ( const auto & child : this->m_children ) {
		alphabet.insert ( child.first );
		child.second->computeMinimalAlphabet ( alphabet );
	}
}

SuffixTrieNodeFinalMark::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace indexes */
