/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProcessor.hpp"
#include <sstream>

namespace measurements {

MeasurementProvisionerResults MeasurementProcessor::process ( const MeasurementProvisionerResults & mpr, const MeasurementProcessorFilter & filter ) {
	MeasurementProvisionerResults newmpr;

	for ( const MPRInputResult & mprir : mpr.inputResults ) {
		MPRInputResult newmprir;
		newmprir.inputs = mprir.inputs;

		for ( const MPRPipelineResult & mprpr : mprir.pipelineResults ) {
			MPRPipelineResult newmprpr;
			newmprpr.pipelineStatus = mprpr.pipelineStatus;

			for ( const MPRPipelineCommandResult & mprpcr : mprpr.commandResults ) {

				 // filter out commands
				if ( !filter.matches ( MeasurementProcessorFilter::FilterType::COMMAND, mprpcr.command ) ) continue;

				MPRPipelineCommandResult newmprpcr;
				newmprpcr.command = mprpcr.command;

				newmprpcr.measurementResults = processMeasurementResults ( mprpcr.measurementResults, filter );

				if ( newmprpcr.measurementResults.frames.size ( ) > 1 )
					newmprpr.commandResults.push_back ( std::move ( newmprpcr ) );
			}

			if ( newmprpr.commandResults.size ( ) )
				newmprir.pipelineResults.push_back ( std::move ( newmprpr ) );
		}

		if ( newmprir.pipelineResults.size ( ) )
			newmpr.inputResults.push_back ( std::move ( newmprir ) );
	}

	return newmpr;
}

MeasurementResults MeasurementProcessor::processMeasurementResults ( const MeasurementResults & mr, const MeasurementProcessorFilter & filter ) {
	MeasurementResults newmr;

	std::function < void ( unsigned, unsigned ) > dfsLambda;

	dfsLambda = [&] ( unsigned idx, unsigned newmrParentIdx ) {
		if ( idx == 0 ) {
			newmr.frames.push_back ( mr.frames[0] );
			newmr.frames[0].subIdxs.clear ( );

			for ( unsigned subIdx : mr.frames[idx].subIdxs )
				dfsLambda ( subIdx, 0 );
		} else {
			const MeasurementFrame & curFrame = mr.frames[idx];

			bool filteredOut = false;
			filteredOut |= !filter.matches ( MeasurementProcessorFilter::FilterType::FRAME_TYPE, to_string ( curFrame.type ) );
			filteredOut |= !filter.matches ( MeasurementProcessorFilter::FilterType::FRAME_NAME, to_string ( curFrame.name ) );

			if ( !filteredOut ) {
				unsigned newmrIdx = newmr.frames.size ( );
				newmr.frames.push_back ( mr.frames[idx] );
				newmr.frames[newmrIdx].subIdxs.clear ( );
				newmr.frames[newmrParentIdx].subIdxs.push_back ( newmrIdx );

				for ( unsigned subIdx : mr.frames[idx].subIdxs )
					dfsLambda ( subIdx, newmrIdx );
			} else {
				for ( unsigned subIdx : mr.frames[idx].subIdxs )
					dfsLambda ( subIdx, newmrParentIdx );
			}
		}
	};

	dfsLambda ( 0, 0 );

	return newmr;
}

MeasurementProvisionerResults MeasurementProcessor::aggregate ( const MeasurementProvisionerResults & mpr ) {

	MeasurementProvisionerResults aggregatedResults;

	std::map < std::vector < std::pair < int, std::string > >, std::vector < MPRInputResult > > sameInputCombinations;

	for ( const auto & mprir : mpr.inputResults )
		sameInputCombinations[mprir.inputs].push_back ( mprir );

	for ( auto & p : sameInputCombinations ) {
		// create one aggregated MPRInputResult

		std::vector < std::vector < MPRPipelineResult > > translatedPipelineResults ( p.second[0].pipelineResults.size ( ) );

		for ( MPRInputResult & mprir : p.second )
			for ( unsigned i = 0; i < translatedPipelineResults.size ( ); ++i )
				translatedPipelineResults[i].push_back ( std::move ( mprir.pipelineResults[i] ) );

		MPRInputResult aggregatedInputResult {
			p.first, std::vector < MPRPipelineResult > ( )
		};

		 // aggregate vector of pipelineResults
		for ( std::vector < MPRPipelineResult > & pipelineResults : translatedPipelineResults ) {
			 // create one aggregated MPRPipelineResult
			std::vector < std::vector < MPRPipelineCommandResult > > translatedPipelineCommandResults ( pipelineResults[0].commandResults.size ( ) );

			for ( MPRPipelineResult & pipelineResult : pipelineResults ) {
				 // filter failed pipelines
				if ( pipelineResult.pipelineStatus.exitCode != 0 ) continue;

				for ( unsigned i = 0; i < translatedPipelineCommandResults.size ( ); ++i )
					translatedPipelineCommandResults[i].push_back ( std::move ( pipelineResult.commandResults[i] ) );
			}

			MPRPipelineResult aggregatedPipelineResult {
				std::move ( pipelineResults[0] ).pipelineStatus, std::vector < MPRPipelineCommandResult > ( )
			};

			 // aggregate vector of commandResults
			for ( std::vector < MPRPipelineCommandResult > & commandResults : translatedPipelineCommandResults ) {
				// create one aggregated MPRPipelineCommandResult

				std::vector < MeasurementResults > measurementResults;

				for ( MPRPipelineCommandResult commandResult : commandResults )
					measurementResults.push_back ( std::move ( commandResult.measurementResults ) );

				MeasurementResults aggregatedMeasurementResults = measurements::MeasurementResults::aggregate ( measurementResults );

				MPRPipelineCommandResult aggregatedCommandResults {
					std::move ( commandResults[0].command ), std::move ( aggregatedMeasurementResults )
				};

				aggregatedPipelineResult.commandResults.push_back ( std::move ( aggregatedCommandResults ) );
			}

			aggregatedInputResult.pipelineResults.push_back ( std::move ( aggregatedPipelineResult ) );
		}

		aggregatedResults.inputResults.push_back ( std::move ( aggregatedInputResult ) );
	}

	return aggregatedResults;
}

std::string MeasurementProcessor::output ( const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & output ) {
	std::stringstream ss;

	MeasurementProcessor::output ( ss, mpr, output );

	return ss.str ( );
}

void MeasurementProcessor::output ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & output ) {
	switch ( output.outputFormat ) {
	case MeasurementProcessorOutput::OutputFormat::XML:
		MeasurementProcessorOutput::outputXml ( os, mpr, output );
		break;

	case MeasurementProcessorOutput::OutputFormat::STATS:
		MeasurementProcessorOutput::outputStats ( os, mpr, output );
		break;

	case MeasurementProcessorOutput::OutputFormat::CSV:
		MeasurementProcessorOutput::outputCsv ( os, mpr, output );
		break;

	case MeasurementProcessorOutput::OutputFormat::HTML:
		MeasurementProcessorOutput::outputHtml ( os, mpr, output );
		break;
	}
}

}
